from pathlib import Path
from datetime import datetime, timedelta
from time import sleep
import os, json
import pandas as pd
from gwpy.time import from_gps


# Function to create symbolic link of the status page into public html
def create_symlink(source: Path, destination: Path) -> None:
    if destination.is_symlink():
        print(f"Symlink already exists at {destination}")
    else:
        try:
            destination.symlink_to(source)
            print(f"Symlink created at {destination}")
        except OSError as e:
            print(f"Creation of the symlink at {destination} failed: {e}")

def find_json_file(directory):

    files = os.listdir(directory)
    for file in files:
        if ".json" in file:
            jsonfile = file
            break

    return directory+'/'+jsonfile

def make_clickable(url, name):
    return '<a href="{}" rel="noopener noreferrer" target="_blank">{}</a>'.format(url,name)

def create_event_table(search_path,format='html',):

    pd_table = None

    triggers_directory = f"{search_path}/trigger_directory"

    events = os.listdir(triggers_directory)
    if len(events)!=0:
        for event_dir in events:
            
            event_name = event_dir.split('-')[0]
            json_file_path = find_json_file(triggers_directory+'/'+event_dir)

            with open(json_file_path) as json_file:
                event_info = json.load(json_file)


            event_info['datetime'] = str(from_gps(event_info['central_time'] ))

            del event_info['scores']
            del event_info['gpstime']

            file_prefix = json_file_path.split('/')[-1][:-5]

            event_info['ifos'] = ''.join(event_info['ifos'])
            event_info['far'] = f"{event_info['far']:.4e}"
            event_info['SNR'] = f"{event_info['SNR']:.2f}"
            event_info['central_time'] = f"{event_info['central_time']:.3f}"
            event_info['central_freq'] = f"{event_info['central_freq']:.1f} Hz"
            event_info['duration'] = f"{event_info['duration']:.5f} sec"
            event_info['bandwidth'] = f"{event_info['bandwidth']:.1f} Hz"


            event_info['GraceDB_ID'] = event_name
            event_info['GraceDB_ID_url'] = f"https://gracedb-test.ligo.org/events/{event_name}/view/"

            event_info['strain_plot_url'] = f"trigger_directory/{event_dir}/{file_prefix}_strain.png"
            event_info['tf_map_url'] = f"trigger_directory/{event_dir}/{file_prefix}_tfmap.png"
            event_info['correlation_plot_url'] = f"trigger_directory/{event_dir}/{file_prefix}_correlation.png"
            event_info['skymap_url'] = f"trigger_directory/{event_dir}/{file_prefix}_skymap.png"

            event_info['Timeseries'] = "link"
            event_info['TFmap'] = "link"
            event_info['Correlation'] = "link"
            event_info['Skymap'] = "link"

            columns = list(event_info.keys())
            columns.insert(0, columns.pop(columns.index('datetime')))
            columns.insert(0, columns.pop(columns.index('GraceDB_ID')))



            if pd_table is None:

                pd_table = pd.DataFrame(event_info,index=[0])
            
            else:
                event_info = pd.DataFrame(event_info,index=[0])
                pd_table = pd.concat([pd_table,event_info],ignore_index=True)
                

        pd_table= pd_table[columns]
        pd_table= pd_table.sort_values(by='central_time',ignore_index=True,ascending=False)

        pd_table['GraceDB_ID'] = pd_table.apply(lambda x: make_clickable(x['GraceDB_ID_url'], x['GraceDB_ID']), axis=1)
        pd_table['Timeseries'] = pd_table.apply(lambda x: make_clickable(x['strain_plot_url'], x['Timeseries']), axis=1)
        pd_table['Correlation'] = pd_table.apply(lambda x: make_clickable(x['correlation_plot_url'], x['Correlation']), axis=1)
        pd_table['TFmap'] = pd_table.apply(lambda x: make_clickable(x['tf_map_url'], x['TFmap']), axis=1)
        pd_table['Skymap'] = pd_table.apply(lambda x: make_clickable(x['skymap_url'], x['Skymap']), axis=1)


        del pd_table['GraceDB_ID_url']
        del pd_table['strain_plot_url']
        del pd_table['correlation_plot_url']
        del pd_table['tf_map_url']
        del pd_table['skymap_url']

        pd_table= pd_table.set_index('GraceDB_ID')

        if format=='html':
            pd_table = pd_table.style.to_html(index=False,classes='center')

    else:

        pd_table = "No events found"

    return pd_table

if __name__ == "__main__":


    # Loading config
    with open('config.json') as json_file:
        config = json.load(json_file)

    # Updating page time 
    status_check_interval = timedelta(seconds=10.0)
    status_timeout = timedelta(seconds=60.0)

    # These two will stay the same for now
    search_path = Path(config['path']) 
    status_page_path = Path(config['path']) / "status.html"

    search_directory_name =search_path.name

    search_symlink_path =  Path(f"/home/{config['user_name']}/public_html/{search_path.name}")

    # Create symlink from search directory to status page directory:
    create_symlink(search_path, search_symlink_path)

    canary_file_path = Path(f"{search_path}/log/search.log")
    seg_plot_path = Path(f"segment_plot.png")
    far_plot_path = Path(f"far.png")
    eff_plot_path = Path(f"Efficiencies.png")

    event_table = create_event_table(search_path=config['path'])


    # the html code which will go in the file GFG.html
    html_template = \
    f"""
    <!DOCTYPE html>
    <html>
    <head>
        <title>MLy-Pipeline Status Page</title>
        <script>
        
            // Check status
            var monitorStarted = new Date();
            
            function updateLastModifiedTime() {{
                // Load the file using XMLHttpRequest
                var xhr = new XMLHttpRequest();
                xhr.open('HEAD', './{canary_file_path}', true);
                xhr.onreadystatechange = function() {{
                    if (xhr.readyState === 4) {{
                        // Get the last modified time of the file
                        var fileLastModified = new Date(xhr.getResponseHeader('Last-Modified'));

                        // Get the current time
                        var currentTime = new Date();
                        
                        console.log(fileLastModified)
                        console.log(currentTime)

                        // Calculate the difference in seconds between the two times
                        var timeDiffInSeconds = (currentTime - fileLastModified) / 1000;

                        // Check if more than timeout seconds have passed since the last modification time
                        if (timeDiffInSeconds > {status_timeout.seconds}) {{
                            document.getElementById('current-status').innerHTML = "Offline since " + monitorStarted;
                            monitorStarted = new Date();
                        }} else {{
                            document.getElementById('current-status').innerHTML = "Online since " + monitorStarted;
                        }}
                        
                        // Update log last modified time
                        document.getElementById('log-last-modified').innerHTML = fileLastModified;
                    }}
                }};
                xhr.send();
            }}
            
            // create a new timestamp     
            
            function refreshImage(){{   
                // create a new timestamp 
                var timestamp = new Date().getTime();  

                var el = document.getElementById("far-plot");  

                var queryString = "?t=" + timestamp;    

                el.src = "{far_plot_path}";    
            }}
            
            
            updateLastModifiedTime();
            
            // Update the last modified time every check miliseconds
            setInterval(updateLastModifiedTime, {status_check_interval.seconds * 1000});
            setInterval(refreshImage, {status_check_interval.seconds * 1000});

        </script>
        </head>
        <body>
            <h1>MLy-Pipeline Status Page</h1>
            <!-- <h2>Search directory: {config['path']}</h2> -->
            <!-- <p>Status: <span id="current-status"></span></p> -->
            <!-- <p>Log Last Modified: <span id="log-last-modified"></span></p> -->
            <h2>Processed segments over available segments</h2>
            <img src="{seg_plot_path}" alt="Segments" id = "seg-plot">

            <h2>Inverse False Alarm Rate</h2>
            <img src="{far_plot_path}" alt="False Alarm Rate" id = "far-plot">

            <h2>Efficiency</h2>
            <img src="{eff_plot_path}" alt="Efficiency" id = "eff-plot">
            
            <h2>List of triggers</h2>
            <h3>{event_table} </h3>
        </body>
        </html>

    """
    
    # writing the code into the file
    with open(status_page_path, 'w') as f:
        f.write(html_template)



#! /usr/bin/env python3

import os, json, time

# Scipy and numpy env parameters that limit the threads
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"

import logging
from logging import handlers
import subprocess

import numpy as np
import argparse
import sys
import pickle

from lal import gpstime

from tensorflow.keras.models import load_model
from tensorflow.config import threading

threading.set_inter_op_parallelism_threads(1)
threading.set_intra_op_parallelism_threads(1)


from mly.datatools import DataPod
from mly.validators import Validator

from .search_functions import *


os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"


def main(**kwargs):
    """ Main functionality of the search script. Given the arguments
    provided it conducts a real time search on gravitational wave data. It
    uses a predefined given model for iference.

    Parameters
    ----------
    
    time_reference: float
        A unix time reference so that should be the same among different functions when
        used splitter. It is suggested to use `unixtime=$(date +%s)` at the beggining of
        the script and pass `$unixtime` as time_reference.    

    splitter: list
        A way to split the search into different parallel jobs. The list must have
        two parameters. The first is the amount of scripts this search is split into.
        The second is which part of the split this function will run. For example a 
        splitter value of [4,1] means that we will splitted the search into 4 scripts
        where each script processes every other 4 seconds, and that this function will
        run the every 4 plus 1 seconds part. For a full search we will have to run the
        same function for splitter values of: [4,0], [4,1], [4,2], [4,3]. If not 
        specified it is equal to [1,0]
        


    Note
    ----

    This function doesn't return anything. It runs until it is stoped or until it
    raises an Exception.
    """


    start_time = time.time()


    with open('config.json') as json_file:
        config = json.load(json_file)
    
    config = { **config , **kwargs }

    # # # Command line arguments processing
    
    # Check arguments and set values in config dict:
    config = checkArguments(config)
    
    np.random.seed( config['seed'] + int(config['script_index']))

    log_level = logging.getLevelName(config["log_level"])

    logger = logging.getLogger("logger_for_search")
    logger.setLevel(logging.DEBUG)


    # create console handler and set level to debug
    ch = handlers.TimedRotatingFileHandler('log/search.log', when='H',interval=12, backupCount=1,)
    ch.setLevel(log_level)

    # create formatter
    formatter = logging.Formatter("%(asctime)s %(levelname)s: %(message)s")

    # add formatter to ch
    ch.setFormatter(formatter)

    # add ch to logger
    logger.addHandler(ch)
    logger.debug(f"PROCESSES after  loading config: {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']]) }")



    # # # Setup parameters for the models
    
    # Loading models
    
    model1 = load_model(config["model1_path"]) # Load coincidence model
    logger.debug(f"PROCESSES after  loading core model1: {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']]) }")

    model2 = load_model(config["model2_path"]) # Load coherence model
    logger.debug(f"PROCESSES after  loading core model2: {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']]) }")

    # logger.info(f"S{config['script_index']} - Loading models complete.")
    # logger.debug(f"PROCESSES after  loading pe models: {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']]) }")

    # This parameter format is used in inference
    
    models = [
        [model1 , model2],
        [["strain"], ["strain","correlation"]]
        ]      
    
    # Mapping keeps consistent the categorical format
    mapping = 2 * [{
        "noise": [1, 0],
        "signal": [0, 1]
    }]
    
    # # # Initialization before loop
    
    # For all scripts we need a common reference of gps time,
    # That time depends on the tiem reference provided at the start
    # and the start lag.
    initial_gps_time = config['time_reference'] - config["start_lag"] 
    
    # GPS index counts how many gps times have been searched by this script alone:
    gps_index = 0 
    
    subconfig = config.copy()
    for key in ['detectors','fs','duration','restriction','prefixes','frames','channels','size']:
        subconfig.pop(key,None)
    
    logger.debug(f"PROCESSES before main loop: {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']]) }")

    # # # Main Loop
    while (1):  
        
        # Get GPS time for this script to search
        gps_time, gps_index = calculateRequiredGPSTime(config, gps_index, initial_gps_time)
        logger.info(f"S{config['script_index']} - Requested / Difference from current GPS time: {gps_time} / {gpstime.gps_time_now() - gps_time}")

        
        loopt0 = time.time()  #Get time at iteration start.
        
        # # Data Aquisition 
        buffers = aquireData(config, gps_time)
        
        getdatatime=time.time()
        
        # If data aquisition fails, reset loop:
        if (buffers is None):
            sys.stdout.flush()
            continue
            
        logger.info(f"S{config['script_index']} - Detector data aquisition time: {getdatatime - loopt0}") 
        
        gpsBeforeInference = gpstime.gps_time_now()
        timeBeforeInference = time.time()
                
        # # Data processing and inference
        
        # Putting data inside datapod format 
        buffer_pod = DataPod(
            np.array(buffers),
            detectors = config["detectors"], 
            fs = config["fs"], 
            gps = len(config["detectors"])*[float(gps_time+0.5)] 
        )
        

        logger.debug(f"PROCESSES before validation: {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']]) }")


        result, thepod = Validator.falseAlarmTest(
                                        models = models,
                                        duration = config["duration"],
                                        fs =  config["fs"],
                                        size = 1,
                                        detectors = config["detectors"],
                                        backgroundType = "real",
                                        noiseSourceFile = buffer_pod,
                                        windowSize = config["required_buffer"],          
                                        mapping = mapping,
                                        strides = None,
                                        plugins = ["correlation_30",'psd','uwstrain'],
                                        restriction = None,
                                        podreturn = True,
                                        **subconfig
                                    )
        
        logger.debug(f"PROCESSES after validation: {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']]) }")

        # Change format of detector name to include 1 (ex. H->H1)
        ifos = [f"{detector}1" for detector in config["detectors"]] 
        
        # Create mly_output dictionary, GraceDB event file basis
        
        interpolated_FAR_t0=time.time()
        interpolated_FAR_trials = 0
        

        while interpolated_FAR_trials < 3:
            try:
                interpolated_FAR = far(config['farfile']+"/"+"FARfile_interpolation.pkl"
                                    ,result["total"][0], inverse = False)
                break
            except:
                try:
                    interpolated_FAR = far(config['farfile']+"/"+"FARfile_interpolation_reserve.pkl"
                                        ,result["total"][0], inverse = False)
                    break
                except Exception as e:
                    interpolated_FAR_trials+=1
                    if interpolated_FAR_trials >= 3:
                        raise e
        logger.info(f"S{config['script_index']} - Loading interpolated_FAR time: ,{time.time()-interpolated_FAR_t0}, {interpolated_FAR_trials}")
        logger.debug(f"PROCESSES after interpolation loading: {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']]) }")

        instance_start_time = result["GPSH"][0]
        mly_output = {
            "gpstime": instance_start_time,
            "far": interpolated_FAR,
            "ifos": ifos,
            "scores": {
                "coincidence": result["scores1"][0],
                "coherency": result["scores2"][0],
                "combined": result["total"][0]
            }

        }
        
        
        # Internal latency, from GPS selection to scores.
        logger.debug(f"S{config['script_index']} - internal_latency {gpsBeforeInference - gps_time}")

        # # # Follow up
        
        # Print scores
        #print(mly_output["scores"])

        # Create string of joined detector names
        detectors = "".join(config["detectors"])

        threshold_t0=time.time()
        threshold_trials = 0
        while threshold_trials < 3:
            try:
                threshold = far(config['farfile']+"/"+"FARfile_interpolation_inverse.pkl"
                                    ,config['far_config']['threshold'], inverse = True)
                break
            except:
                try:
                    threshold = far(config['farfile']+"/"+"FARfile_interpolation_inverse_reserve.pkl"
                                        ,config['far_config']['threshold'], inverse = True)
                    break
                except Exception as e:
                    threshold_trials+=1
                    if threshold_trials >= 3:
                        raise e
        logger.debug(f"S{config['script_index']} - Loading threshold interpolation time: {time.time()-threshold_t0,threshold_trials}")

        # Perform time- and frequency-domain parameter estimation.
        # It adds parameter estimation to events.
        pe_t0 = time.time()
        mly_output = runTimeFrequencyParameterEstimation(thepod,  mly_output)
        logger.info(f"PE time {time.time()- pe_t0}s")
        # If result us above threshold, do all even relate actions 
                
        logger.debug(f"PROCESSES after runTimeFrequencyParameterEstimation : {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']]) }")

        if mly_output["scores"]["combined"] >= threshold:

            eventDirectory = 'tempEventDirectory_'+str(mly_output['gpstime'])
            os.mkdir(f"{config['trigger_directory']}/{eventDirectory}")

            # Saving trigger into the trigger_directory
            with open(f"{config['trigger_directory']}/{eventDirectory}/T_{mly_output['gpstime']}_{detectors}.json", "w") as mly_json:
                json.dump(mly_output, mly_json,indent=4)
                mly_json.close()

            # Saving trigger DataPod into the trigger_directory
            with open(f"{config['trigger_directory']}/{eventDirectory}/T_{mly_output['gpstime']}_{detectors}.pkl", 'wb') as mly_pkl:
                pickle.dump(thepod, mly_pkl, 4)
                
            # Saving trigger background DataPod into the trigger_directory
            with open(f"{config['trigger_directory']}/{eventDirectory}/T_{mly_output['gpstime']}_{detectors}_buffer.pkl", 'wb') as mly_pkl:
                pickle.dump(buffer_pod, mly_pkl, 4)
                                
            # Sending trigger to GraceDB 
            logger.debug(f"PROCESSES before run mlytograce: {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']])}")

            with open( f"{config['trigger_directory']}/{eventDirectory}/T_{mly_output['gpstime']}_{detectors}.out"
            ,"wb") as out:
                # Using subprocess to create mly_to grace function and run it independently.
                p = subprocess.Popen(["python","-m","mly_pipeline.mly_to_grace", "--triggerfile"
                                      ,f"{config['trigger_directory']}/{eventDirectory}/T_{mly_output['gpstime']}_{detectors}.json"
                                      ,"--trigger_destination", str(config['trigger_destination'])]
                                      ,stdout=out, stderr=out,shell=False)
            
            logger.info(f"S{config['script_index']} - TRIGGER follow-up script initialised for: "+str(mly_output['gpstime']))
                
            # Saving output to output directory. Notice that the name starts with T_ for triggers
            if config["output_directory"] != None:
                with open(f"{config['output_directory']}T_{mly_output['gpstime']}_{detectors}.json", "w") as mly_json:
                    json.dump(mly_output, mly_json,indent=4)
                    mly_json.close()
            
            logger.debug(f"PROCESSES after run mlytograce: {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']])}")

        # If result is below threshold, we just save the output. Note that the name starts with N_
        elif config["output_directory"] != None:
            
            with open(f"{config['output_directory']}N_{instance_start_time}_{detectors}.json", "w") as mly_json:
                json.dump(mly_output, mly_json, indent=4)
                mly_json.close()
        
        logger.debug(f"PROCESSES after saving: {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']]) }")

        # If fileSystem parameter is provided, we save the second processed by the model in
        # the file system directory.
        if config["masterDirectory"]!=None:
            podToFileSystem(thepod, masterDirectory = config["masterDirectory"])

        # If bufferDirectory is provided, we save the buffer_pod in the 
        # buffer directory.
        if config["bufferDirectory"] != None:
            buffer_pod.save(config["bufferDirectory"] 
                            + "temp/" + str(buffer_pod.gps[0]) 
                            + "_" + str(int(buffer_pod.duration)))  
        
        # Timing the loop to see if we need more splits in splitter.
        print(f"Processing loop duration: {str(time.time()-timeBeforeInference)}")  
        
        gps_time += config["num_scripts"]
        
        sys.stdout.flush()

        if time.time() - start_time > 24*3600:

            os.system(f"> search_step_{config['script_index']}.out")      
        


if __name__ == "__main__":

    # List of arguments to pass:
    arguments = [
        "splitter",
        "time_reference",
    ]

    # Construct argument parser:
    parser = argparse.ArgumentParser()
    [parser.add_argument(f"--{argument}") for argument in arguments]

    # Pass arguments:
    args = parser.parse_args()

    # Store arguments in dictionary:
    kwargs = {}
    for argument in arguments:
        kwargs[argument] = getattr(args, argument)

    main(**kwargs)



#! /usr/bin/env python
import subprocess, argparse, os, shutil, gwdatafind
import time, sys ,pickle, json, datetime, math

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import astropy_healpix as ah
import healpy as hp

from gwpy.timeseries import TimeSeries
from ligo.skymap.io import fits
from ligo.skymap.moc import nest2uniq 
from mly.skymap_utils import skymap_plugin, containment_region
from pycondor import Job, Dagman

from mly.datatools import DataPod
from mly.validators import Validator
from mly.createFileSystem import createFileSysem
from mly.plugins import PlugIn
from .search_functions import *

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
os.environ["GWDATAFIND_SERVER"]="datafind.ldas.cit:80"
os.environ["DEFAULT_SEGMENT_SERVER"]="http://segments.ldas.cit"

from tensorflow.keras.models import load_model
from tensorflow.config import threading

threading.set_inter_op_parallelism_threads(1)
threading.set_intra_op_parallelism_threads(1)

which_python = str(subprocess.check_output(['which','python']))[2:-3]

def main(**kwargs):
    """ Main functionality of the search script. Given the arguments
    provided it conducts a real time search on gravitational wave data. It
    uses a predefined given model for iference.

    Parameters
    ----------

    detectors: {'HL','HLV'}
        It specifies which detectors to use to conduct the search. If 'HL' is
        chosen, it generates gaussian noise that follows Virgo background(x32),
        to provide nominal data for Virgo.

    channels: {'NOISE','MDC'}
        It specifies which channel to use for the search.
        'NOISE' represents a predefined set of channels without injections.
        'MDC' represents a predifned set of channels with injections for the MDC

    threshold : float
        The minimum score for witch the search will issue a trigger event.

    output_directory: str (path, optional)
        The path of the directory where the search will save the output of the model
        for each instance processed. If not specified it doesn't save the outputs.

    trigger_directory: str (path, optional)
        The path of the directory where the search will save the output of the model
        for each instance that was above the threshold value. If not specified
        it doesn't save anything. It is important to have this directory specified if
        you want to send triggers to GraceDB (using mly_to_grace.py).

    trigger_destination: {'test, 'playground', 'dev1', None}
        The GraceDB domain for the triggers to be uploaded. Each option represent
        a corresponding url. If equals to None, it will not issue an GraceDB event.
        
    splitter: list
        A way to split the search into different parallel jobs. The list must have
        two parameters. The first is the amount of scripts this search is split into.
        The second is which part of the split this function will run. For example a 
        splitter value of [4,1] means that we will splitted the search into 4 scripts
        where each script processes every other 4 seconds, and that this function will
        run the every 4 plus 1 seconds part. For a full search we will have to run the
        same function for splitter values of: [4,0], [4,1], [4,2], [4,3]. If not 
        specified it is equal to [1,0]
        
    skymap: bool
        If True it will allow the creation of a skymap whenever it issues events. This
        is passed to another function 'mly_to_grace'.
    
    time_reference: float
        A unix time reference so that should be the same among different functions when
        used splitter. It is suggested to use `unixtime=$(date +%s)` at the beggining of
        the script and pass `$unixtime` as time_reference.    
        
    fileSystem: str/path
        A path to a valid file system directory. A valid file system
        directory needs to have subfolders with the initials of all
        detectors used and a 'temp' file that also includes subfolders
        with initials of all detectors. This is used for the calculation
        of continues false alarm rates. If not specified, it will not save data 
        into the file system.
        
    bufferDirectory: str/path
        A directory path where it will save the buffer data around each second
        to be used for the calculation of efficiencies. If not specified, it 
        will not save the buffer data.
        
    Note
    ----

    This function doesn't return anything. It runs until it is stoped or until it
    raises an Exception.
    """
    
    with open('config.json') as json_file:
        config = json.load(json_file)
    
    config = { **config , **kwargs }
    # # # Command line arguments processing
    
    # Check arguments and set values in config dict:
    config = checkArguments(config)

    # # # Setup parameters for the models
    
    subconfig = config.copy()
    for key in ['detectors','fs','duration','restriction','prefixes','frames','channels','size']:
        subconfig.pop(key,None)
    
    detectors = config["detectors"]
    prefix = config['prefix']
    frames_dir = config['frames_directory']
    channels = config['channels']
    segment_list = config['segment_list']

    if kwargs['mode'] == 'condor':

        error = './condor/error'
        output = './condor/output'
        log = './condor/log'
        submit = './condor/submit'

        dagman = Dagman(name=config['path'].split('/')[-1]+'_'+'offline_search',submit=submit)

        # # # Main Loop

        
        for i, segment in enumerate(segment_list):

            start_gps = int(segment[0])
            segment_size = int(segment[1] - segment[0])
            file_name_prefix = prefix + str(start_gps) +"-"+ str(segment_size)

            jobname = file_name_prefix+"_"+str(i)

            if os.path.isfile(config['output_directory']+ file_name_prefix +".pkl"):
                with open(config['output_directory']+ file_name_prefix +".pkl",'rb') as out:
                    out_ = pickle.load(out) 

                print("Segment " + file_name_prefix + " already processed. Expected size: ", segment_size, ". Finall size: " ,len(out_))
                
                continue


    
            thearguments = ("-m mly_pipeline.offline_search --mode job "+"--i "+str(i))

            job = Job(name = jobname
                    ,executable = which_python
                    ,arguments = thearguments
                    ,submit=submit
                    ,error=error
                    ,output=output
                    ,log=log
                    ,getenv=True
                    ,dag=dagman
                    ,extra_lines=["accounting_group_user="+config['accounting_group_user']
                                    ,"accounting_group="+"ligo.dev.o4.burst.allsky.mlyoffline"
                                    ,"request_disk            = 204800"
                                    ,"request_memory = 8192"
                                    ])


        dagman.build_submit()
        with open('monitor.log', 'a') as file:
            # Write the line to the file
            line_to_add = f"{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')} Offline search - Number of jobs submited: {len(segment_list)}\n"
            file.write(line_to_add)


    elif kwargs['mode']=='job':

        i = int(kwargs['i'])

        np.random.seed(config['seed'] + i)

        segment = segment_list[i]

        # Loading models
        model1 = load_model(config["model1_path"]) # Load coincidence model
        model2 = load_model(config["model2_path"]) # Load coherence model

        # This parameter format is used in inference
        models = [
            [model1 , model2],
            [["strain"], ["strain","correlation"]]
            ]      
        
        # Mapping keeps consistent the categorical format
        mapping = 2 * [{
            "noise": [1, 0],
            "signal": [0, 1]
        }]


        # # # Main Loop

        # THE SEGMENTS GUARANTEE THAT THE TIMES ARE OF COINSIDECE

        background_data = []

        start_gps = int(segment[0])
        segment_size = int(segment[1] - segment[0])
        file_name_prefix = prefix + str(start_gps) +"-"+ str(segment_size)

        for det in detectors:
            
            if config['prefix']=='noprefix':

                print('seg',det,config['frames_directory'][det],segment[0],segment[1])
                conn=gwdatafind.connect()
                urls=conn.find_urls(det
                                , config['frames_directory'][det]
                                , segment[0]
                                , segment[1])
                                
                data = TimeSeries.read(urls
                                , config['channels'][det]
                                , start = segment[0]
                                , end = segment[1]
                                ).resample(config["fs"]).astype('float64').value

                print("background data ",det,data.shape)
                background_data.append(data)
                                       
            else:

            
                data = TimeSeries.read([frames_dir[det]+ file  for file in os.listdir(frames_dir[det]) if det+'-'+det+"1" in file]
                                        , channels[det]
                                        , start = segment_list[i][0]
                                        , end = segment_list[i][1])
                data = data.resample(config["fs"]).value
                print("background data ",det,data.shape)
                background_data.append(data)
            


        mly_output_list = []

        for j in range(int(segment_size/config["duration"] - (config["required_buffer"] - config["duration"]) )):

            gps_time = start_gps + j

            strain_data_j = np.array(background_data)[: , j*config["fs"] : (j+config["required_buffer"])*config["fs"] ]
            
            zeros=len(np.where(strain_data_j==0.0)[0]) # weidly it returns array in a tuple
            if zeros>=0.10*config["required_buffer"]*config["fs"]:
                print(f"Error! Data has many zeros, skipped to the next.")
                continue

            
            buffer_pod = DataPod(np.array(background_data)[: , j*config["fs"] : (j+config["required_buffer"])*config["fs"] ],
                        detectors = config["detectors"], 
                        fs = config["fs"], 
                        gps = len(config["detectors"])*[float(gps_time)] 
                            )

            try:
                result, thepod = Validator.falseAlarmTest(
                                                models = models,
                                                duration = config["duration"],
                                                fs =  config["fs"],
                                                size = 1,
                                                detectors = config["detectors"],
                                                backgroundType = "real",
                                                noiseSourceFile = buffer_pod,
                                                windowSize = config["required_buffer"],          
                                                mapping = mapping,
                                                strides = None,
                                                plugins = ["correlation_30",'psd','uwstrain'],
                                                restriction = None,
                                                podreturn = True,
                                                **subconfig
                                            )
            except Exception as e:
                raise(e)
                continue
            # Change format of detector name to include 1 (ex. H->H1)
            ifos = [f"{detector}1" for detector in config["detectors"]] 
            
            # Create mly_output dictionary, GraceDB event file basis
            mly_output = {
                "gpstime": result["GPSH"][0],
                "far": far(config['farfile']+"/"+"FARfile_interpolation.pkl"
                                            ,result["total"][0], inverse = False),
                "ifos": ifos,
                "scores": {
                    "coincidence": result["scores1"][0],
                    "coherency": result["scores2"][0],
                    "combined": result["total"][0]
                }

            }


            # Internal latency, from GPS selection to scores.

            # # # Follow up

            # Create string of joined detector names
            detectors = "".join(config["detectors"])

            
            threshold = far(config['farfile']+"/"+"FARfile_interpolation_inverse.pkl"
                                ,config['far_config']['threshold'], inverse = True)
            # If result us above threshold, do all even relate actions 

            # Perform time- and frequency-domain parameter estimation.
            # It adds parameter estimation to events.

            mly_output = runTimeFrequencyParameterEstimation(thepod,  mly_output)

            if mly_output["scores"]["combined"] >= threshold:
                
                mly_output['trigger']=True
                
                # Making the directory of the trigger
                eventDirectory = "T_"+str(mly_output['gpstime'])

                if os.path.exists(f"{config['trigger_directory']}/{eventDirectory}"):
                    shutil.rmtree(f"{config['trigger_directory']}/{eventDirectory}")
                
                os.mkdir(f"{config['trigger_directory']}/{eventDirectory}")


                # Saving trigger into the trigger_directory
                with open(f"{config['trigger_directory']}/{eventDirectory}/T_{mly_output['gpstime']}_{detectors}.json", "w") as mly_json:
                    json.dump(mly_output, mly_json,indent=4)
                    mly_json.close()

                # Saving trigger DataPod into the trigger_directory
                with open(f"{config['trigger_directory']}/{eventDirectory}/T_{mly_output['gpstime']}_{detectors}.pkl", 'wb') as mly_pkl:
                    pickle.dump(thepod, mly_pkl, 4)
                    
                # Saving trigger background DataPod into the trigger_directory
                with open(f"{config['trigger_directory']}/{eventDirectory}/T_{mly_output['gpstime']}_{detectors}_buffer.pkl", 'wb') as mly_pkl:
                    pickle.dump(buffer_pod, mly_pkl, 4)

                # Creating the strain plot
                thepod.plot(type_="strain")
                plt.savefig(f"{config['trigger_directory']}/{eventDirectory}/T_{mly_output['gpstime']}_{detectors}_strain.png")
                plt.clf()
                # Creating the correlation plot
                thepod.plot(type_="correlation")
                plt.savefig(
                    f"{config['trigger_directory']}/{eventDirectory}/T_{mly_output['gpstime']}_{detectors}_correlation.png")
                plt.clf()
                # Creating the tf_map plot
                thepod.plot('tf_map')
                plt.savefig(
                    f"{config['trigger_directory']}/{eventDirectory}/T_{mly_output['gpstime']}_{detectors}_tfmap.png")

                if config['skymap']:

                    if ('V' in thepod.detectors) and ('V' not in config['detectors']):
                        thepod.strain = thepod.strain[:2]
                        thepod.detectors = thepod.detectors[:2]
                        thepod.gps = thepod.gps[:2]
                        thepod.psd = thepod.psd[:2]
                        thepod.uwstrain = thepod.uwstrain[:2]

                    # Create skymap plugin:
                    thepod.addPlugIn(skymap_plugin(alpha = config['skymap']['alpha']
                                                ,beta = config['skymap']['beta']
                                                ,sigma =config['skymap']['sigma']
                                                ,nside = config['skymap']['nside']
                                                ,window_parameter = (thepod.PE['start_time'], thepod.PE['end_time'], 1/16)))
                    
                    skymap_path = f"{config['trigger_directory']}/{eventDirectory}/mly.multiorder.fits"
                    
                    skymap_prob = thepod.sky_map[0]

                    order = int(math.log2(config['skymap']['nside']))
                    npix = ah.nside_to_npix(config['skymap']['nside'])

                    uniq = nest2uniq(np.uint8(order), np.arange(npix))
                    probdensity = skymap_prob / hp.nside2pixarea(config['skymap']['nside'])

                    moc_data = np.rec.fromarrays(
                        [uniq, probdensity], names=['UNIQ', 'PROBDENSITY'])

                    with open(skymap_path, "w") as f:
                        fits.write_sky_map( skymap_path, moc_data)
                    
                    thepod.plot('sky_map')
                    plt.savefig(
                        f"{config['trigger_directory']}/{eventDirectory}/T_{mly_output['gpstime']}_{detectors}_skymap.png")

            else:

                mly_output['trigger']=False

            mly_output_list.append(mly_output)
            sys.stdout.flush()


        tempFrame = pd.DataFrame(columns = ['gpstime','far','ifos'
                                                    ,'coincidence','coherency'
                                                    ,'combined','central_time','duration','central_freq','bandwidth'
                                                    ,'trigger'])


        # A check value to make sure we add all the json files.
        frameSizeCheck = len(mly_output_list) 

        # Adding all the events in json format to the tempFrame
        for file in mly_output_list:

            p=file.copy()
                
            for k in list(p['scores'].keys()):
                p[k]=p['scores'][k]

            del(p['scores'])  

            p['ifos'] = ",".join(p['ifos'])

            for k in list(p.keys()):
                p[k]=[p[k]]


            tempFrame=pd.concat([tempFrame, pd.DataFrame.from_dict(p)]
                                    ,ignore_index=True)

                # Reseting index and sorting by GPS time
            tempFrame = tempFrame.sort_values(by="gpstime").reset_index(drop=True)
            tempFrame = tempFrame[['gpstime','far','ifos'
                                                    ,'coincidence','coherency'
                                                    ,'combined','central_time','duration','central_freq','bandwidth'
                                                    ,'trigger']]


        with open(config['output_directory']+file_name_prefix+".pkl", 'wb') as output:
            pickle.dump(tempFrame, output, 4)


        print(config['output_directory']+file_name_prefix + ' saved')
        print("Size of file and interval ",frameSizeCheck, segment_size)

        print("Analysis complete ",str(1+i),"/",len(file_name_prefix))

    elif kwargs['mode']=='set_file_system':

        np.random.seed(config['seed'] - 1)

        createFileSysem(duration=config['duration']
                         ,fs=config['fs']
                         ,detectors=detectors
                         ,dates = segment_list
                         ,windowSize = 16
                         ,backgroundType='real'
                         ,masterDirectory=config['masterDirectory']
                         ,frames=config['frames_directory'] 
                         ,channels=config['channels']
                         ,observingFlags = config['active_flags']
                         ,accounting_group_user = config['accounting_group_user']
                         ,accounting_group = config['accounting_group']
                         ,dagman_name = f"{config['path'].split('/')[-1]}_createFileSystem"

        )




if __name__ == "__main__":

    # # # Managing arguments
    # 
    # List of arguments to pass:
    arguments = ['mode','i']

    #Construct argument parser:
    parser = argparse.ArgumentParser()

    parser.add_argument('--mode', default="condor")
    parser.add_argument('--i', default = None)
    # Pass arguments
    args = parser.parse_args()

    # Store arguments in dictionary:
    kwargs = {}
    for argument in arguments:
        kwargs[argument] = getattr(args, argument)

    kwargs["splitter"] = None
    kwargs["time_reference"] = 0

    main(**kwargs)



#! /usr/bin/env python3

import os, json, math, time

# Scipy and numpy env parameters that limit the threads
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"

import argparse
import numpy as np
import matplotlib.pyplot as plt

from lal import gpstime
from ligo.gracedb.rest import GraceDb

# # # Managing arguments
# 
# List of arguments to pass:
arguments = [
    "trigger_destination",
    "triggerfile",
]

#Construct argument parser
parser = argparse.ArgumentParser()
[parser.add_argument(f"--{argument}") for argument in arguments]

# Pass arguments
args = parser.parse_args()

# Store arguments in dictionary
kwargs = {}
for argument in arguments:
    kwargs[argument] = getattr(args, argument)

# Config file is inherited by the search script updated
with open('config.json') as json_file:
    config = json.load(json_file)

# config = ast.literal_eval(kwargs['config'])  
    
    
"""This function issues a GraceDB event provided with a json file.

Parameters
----------


trigger_destination: {'test, 'playground', 'dev1', None}
    The GraceDB domain for the triggers to be uploaded. Each option represent
    a corresponding url. If equals to None, it will not issue an GraceDB event.

triggerfile : 'str'
    The path to the json file to use for the event.

"""

# print("Main beggining: ",gpstime.gps_time_now()- float(kwargs['triggerfile'].split('_')[-2]))
# Argument checks

if kwargs['triggerfile']!=None:

    if os.path.isfile(kwargs['triggerfile']):
        triggerfile = kwargs['triggerfile']
    else:
        raise FileNotFoundError(
            kwargs['triggerfile'] +
            " is not a valid file path")
else:
    raise FileNotFoundError("You need to specify the trigger file path.")




# # # Push to GraceDb

# Checking latencies
gpsbefore = gpstime.gps_time_now()

# If the authentication is correct and we can send events, trigger_destination
# can be defined.


if kwargs['trigger_destination'] != 'None':

    url=kwargs['trigger_destination']

    client = GraceDb(service_url=url)

    graceEventOutput = client.create_event(group=config["trigger_group"], 
                                           pipeline="MLy",
                                           filename=triggerfile,
                                           search="AllSky",
                                           labels = ['PE_READY'])
    
    graceEventDict = graceEventOutput.json()
    
    print(triggerfile+" uploaded to "+url)

# Otherwise for just testing we can create a fake id to use.
else:
    
    graceEventDict = {'graceid':'FAKEID'+str(np.random.randint(999999))}

# Checking latency from the GPS time involved    
print("Latency before opening client: ",gpsbefore- float(kwargs['triggerfile'].split('_')[-2]))
timeafterclient = time.time()

# # # Extra features to be added
#
# To this point an event is created. The following is to be added after the
# event creation in the future. Some arguments are checked here so that 
# they don't contribute to the event latency until their feature is used



from mly.datatools import DataPod

# Loading the pod that has all the data that produced the trigger
thepod = DataPod.load(triggerfile[:-4]+'pkl')

gpsstring=triggerfile.split('_')[-2]
detectors=triggerfile.split('.')[-2].split('_')[-1]

# Renaming the already created tempEventDirectory to one with the gracedb id in the name

temp_eventDirectory = 'tempEventDirectory_'+gpsstring
# Creating a directory for each event using the graceid from the event creation.
# Not always works.
try:
    eventDirectory = graceEventDict['graceid']+'-'+gpsstring+'-'+detectors
except Exception as e:
    print(e)
    eventDirectory = 'NoGraceID'+'-'+gpsstring+'-'+detectors

# Renaming the event directory
os.rename(  f"{config['trigger_directory']}/{temp_eventDirectory}"
          , f"{config['trigger_directory']}/{eventDirectory}")

# Creating the strain plot
thepod.plot(type_="strain")
plt.savefig(f"{config['trigger_directory']}/{eventDirectory}/T_{gpsstring}_{detectors}_strain.png")

# Creating the correlation plot
thepod.plot(type_="correlation")
plt.savefig(f"{config['trigger_directory']}/{eventDirectory}/T_{gpsstring}_{detectors}_correlation.png")

thepod.plot('tf_map')
plt.savefig(f"{config['trigger_directory']}/{eventDirectory}/T_{gpsstring}_{detectors}_tfmap.png")

print(config['skymap'], bool(config['skymap']))
if config['skymap']:

    from mly.skymap_utils import skymap_plugin, containment_region
    from mly.plugins import PlugIn
    from ligo.skymap.io import fits
    from ligo.skymap.moc import nest2uniq 
    import astropy_healpix as ah
    import healpy as hp
    from tensorflow.config import threading

    threading.set_inter_op_parallelism_threads(1)
    threading.set_intra_op_parallelism_threads(1)

    if ('V' in thepod.detectors) and ('V' not in config['detectors']):
        thepod.strain = thepod.strain[:2]
        thepod.detectors = thepod.detectors[:2]
        thepod.gps = thepod.gps[:2]
        thepod.psd = thepod.psd[:2]
        thepod.uwstrain = thepod.uwstrain[:2]

    # Create skymap plugin:
    thepod.addPlugIn(skymap_plugin(alpha = config['skymap']['alpha']
                                  ,beta = config['skymap']['beta']
                                  ,sigma =config['skymap']['sigma']
                                  ,nside = config['skymap']['nside']
                                  ,window_parameter = (thepod.PE['start_time'], thepod.PE['end_time'], 1/16)))

    skymap_path = f"{config['trigger_directory']}/{eventDirectory}/mly.multiorder.fits"
    
    skymap_prob = thepod.sky_map[0]

    order = int(math.log2(config['skymap']['nside']))
    npix = ah.nside_to_npix(config['skymap']['nside'])

    uniq = nest2uniq(np.uint8(order), np.arange(npix))
    probdensity = skymap_prob / hp.nside2pixarea(config['skymap']['nside'])

    moc_data = np.rec.fromarrays(
        [uniq, probdensity], names=['UNIQ', 'PROBDENSITY'])

    with open(skymap_path, "w") as f:
        fits.write_sky_map( skymap_path, moc_data)
    
    # Upload skymap to grace db:
    if kwargs['trigger_destination'] != 'None':

        client.write_log(graceEventDict['graceid'], 'Sky Localization',  filename=skymap_path, tg_name='sky_loc')
        client.write_label(graceEventDict['graceid'], 'SKYMAP_READY')
    
    print("Latency after skymap log: ",time.time()-timeafterclient)

    thepod.plot('sky_map')
    plt.savefig(f"{config['trigger_directory']}/{eventDirectory}/T_{gpsstring}_{detectors}_skymap.png")

    if kwargs['trigger_destination'] != 'None':

        plot_string = f"{config['trigger_directory']}/{eventDirectory}/T_{gpsstring}_{detectors}"

        client.write_log(graceEventDict['graceid'], 'TimeSeries Plot',  filename=f"{plot_string}_strain.png", tg_name='strain')
        client.write_log(graceEventDict['graceid'], 'Correlation Plot',  filename=f"{plot_string}_correlation.png", tg_name='strain')
        client.write_log(graceEventDict['graceid'], 'Time-Frequency Map',  filename=f"{plot_string}_tfmap.png", tg_name='tfplots')
        client.write_log(graceEventDict['graceid'], 'Skymap Probability',  filename=f"{plot_string}_skymap.png", tg_name='sky_loc')


raise SystemExit()

# if __name__ == "__main__":


#     main(**kwargs)

#     quit()




import os
# Scipy and numpy env parameters that limit the threads
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"

import json
import pickle
import argparse
import sys ,time
import pandas as pd
import os
import logging
import subprocess
import random

from mly.tools import dirlist
from mly.datatools import *
from .search_functions import *
from dqsegdb2.query import query_segments
from gwpy.segments import Segment, SegmentList, DataQualityFlag
from gwpy.time import from_gps, to_gps
from datetime import datetime


# # # Managing arguments
# 
# List of arguments to pass:
arguments = [

    "restriction",
    "timeUnit"
]

#Construct argument parser:
parser = argparse.ArgumentParser()

parser.add_argument('--restriction')
parser.add_argument('--timeUnit')


# Pass arguments
args = parser.parse_args()

# Store arguments in dictionary:
kwdict = {}
for argument in arguments:
    kwdict[argument] = getattr(args, argument)
    
with open('config.json') as json_file:
    config = json.load(json_file)
    
config={ **kwdict , **config}


# # # Initialising logging
log_level = logging.getLevelName(config["log_level"])

# create logger
logger = logging.getLogger("logger_for_manager")
logger.setLevel(log_level)



# create console handler and set level to debug
ch = logging.handlers.TimedRotatingFileHandler('log/manager.log', when='H',interval=12, backupCount=1,)
ch.setLevel(logging.DEBUG)

# create formatter
formatter = logging.Formatter("%(asctime)s %(levelname)s: %(message)s")

# add formatter to ch
ch.setFormatter(formatter)

# add ch to logger
logger.addHandler(ch)





def main(config):

    config = checkOutputDirectoryArguments(config)

    # The size in which we group the outputs
    if config['maxDataFrameSize']==None: 
        config['maxDataFrameSize']=3600

    maxDataFrameSize = int(config['maxDataFrameSize'])


    # This is the time scale to use for the plots
    timeUnitDict={1:'s', 60:'m', 3600:'h', 24*3600:'days',30*24*3600:'months', 365*24*3600:'years'}

    if config['timeUnit']==None: 
        config['timeUnit']=3600
    else:
        config['timeUnit']=int(config['timeUnit'])
        

    sys.stdout.flush()


    lastdatetimehour = -1 # Hour index so that some processes run every hour only

    while (1):
        logger.debug(f"PROCESSES before output management: {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']]) }")

        loopT0=time.time()
        
        output_directory = config['output_directory']
        
        frameSizeCheck=0
        fileList = dirlist(output_directory)
        dFramesList  = []
        jsonList = []
        # Listing the files in the directory
        # Separating the json from pkl
        for file in fileList:

            if ".json" in file:
                jsonList.append(file)
            elif ".pkl" in file:
                dFramesList.append(file)

        # if there are no new outputs we skip that step
        if len(jsonList)!=0: 

            # We expect that every time the manager runs, there is a tempFrame
            # file that might have events but it is not of the size we group
            # them (maxDataFrameSize). 
            try:

                with open(output_directory+'tempFrame.pkl','rb') as obj:
                    tempFrame = pickle.load(obj)
            # The first time manager is running this file is created instead.
            except Exception as e:
                tempFrame = pd.DataFrame(columns = ['gpstime','far','ifos'
                                                    ,'coincidence','coherency'
                                                    ,'combined','trigger'])
                logger.info(f"Creating tempfile for the first time.")

            # The size of the frame
            initialTempFrameSize=len(tempFrame)

            # A check value to make sure we add all the json files.
            frameSizeCheck = len(jsonList) + initialTempFrameSize

            # Adding all the events in json format to the tempFrame
            for file in jsonList:
                try:
                    with open(output_directory+file,"r") as jf:
                        p=json.load(jf)
                        jf.close()
                except Exception as e:
                    logger.warning(f"Exception in loading json file {file}, {e}")
                    continue



            # Reformating the score parameters to the desired format.
                for k in list(p['scores'].keys()):
                    p[k]=p['scores'][k]
                # Making a list of ifos to avoid line breaking into two
                p['ifos'] = [p['ifos']]

                if file[0]=='N':
                    p['trigger']=False
                elif file[0]=='T':
                    p['trigger']=True
                else:
                    p['trigger']=False


                del(p['scores'])  
                tempFrame=pd.concat([tempFrame, pd.DataFrame.from_dict(p)]
                                    ,ignore_index=True)

            # Reseting index and sorting by GPS time
            tempFrame=tempFrame.sort_values(by="gpstime").reset_index(drop=True)
            # Rearanging the columns
            tempFrame = tempFrame[['gpstime','far','ifos','coincidence','coherency'
                                ,'combined','trigger']]
            # Chopping the tempFrame to chuncs of maxDataFrameSize outputs
            # IMPORTANT: We don't separate them by hour, we just collect 
            # maxDataFrameSize of them together

            _frame = tempFrame[:10]


            while len(tempFrame) > maxDataFrameSize:

                # Creating a newframe to include the maxDataFrameSize instances
                newFrame=tempFrame[:maxDataFrameSize]
                with open(output_directory+str(newFrame['gpstime'][0])+"-"
                        +str(newFrame['gpstime'][maxDataFrameSize-1])
                        +"_"+str(maxDataFrameSize)+'.pkl', 'wb') as output:
                    pickle.dump(newFrame, output, 4)

                # logger.info("Saving group of outputs: ", output_directory+str(newFrame['gpstime'][0])+"-"
                #     +str(newFrame['gpstime'][maxDataFrameSize-1])
                #     +"-"+str(maxDataFrameSize)+'.pkl')

                # Removing the newframe from tempFrame
                tempFrame=tempFrame[maxDataFrameSize:].reset_index(drop=True)

                frameSizeCheck-=maxDataFrameSize

            # What is left in tempFrame, overights the tempFrame.pkl
            with open(output_directory+'tempFrame.pkl', 'wb') as output:
                pickle.dump(tempFrame, output, 4)
                logger.info(f"Saving the incomplete group of outputs: {output_directory}tempFrame.pkl")

            
            # We remove all json files already included in a dataFrame
            for file in jsonList:
                os.system("rm "+output_directory+file )


            
        logger.debug(f"PROCESSES before FAR management: {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']]) }")

        # # # FileSystem for FAR management
        #
        # Listing all files saved in the temp file (files to be managed)
        detectors= list( det for det in config['detectors'])
        
        t0=time.time()

        numberOfFiles={}
        fileNumbers={}
        dataSetDict={}
        
        # For each detector we count how many files have been saved in all
        # detectors and we sort them.
        for det in detectors:
            numberOfFiles[det] = len(dirlist(config['masterDirectory']+'/temp/'+det))
            fileNumbers[det] = sorted(list(file.split('_')[0] for file in dirlist(
                config['masterDirectory']+'/temp/'+det)))
            dataSetDict[det] = []

        # If there are no files we skip this step
        if len(fileNumbers[detectors[0]])==0:
            logger.info(f"Empty temp directory")
            #time.sleep(60)
            
        # If the files included in temp are less than maxDataFrameSize, we skip
        # this step.
        elif (int(float(fileNumbers[detectors[0]][-1]))
            -int(float(fileNumbers[detectors[0]][0]))) < maxDataFrameSize:
            logger.info(f"Not enough data \
                        {maxDataFrameSize - (int(float(fileNumbers[detectors[0]][-1])) -  int(float(fileNumbers[detectors[0]][0])))}\
                        to run the FAR test yet")
            
        # If the files included in temp are enough we group them
        elif (int(float(fileNumbers[detectors[0]][-1]))
            -int(float(fileNumbers[detectors[0]][0]))) >= maxDataFrameSize:
            
            c=0
            
            # We keep track of the files used to delete them later
            filesToDelete=[]
            
            # Checking that all detectors have the same number of files
            logger.debug(f"Number of files for each detector: {list(len(fileNumbers[det]) for det in detectors)}")
            
            lastFileNumber=None
            # print("group size check: ",len(fileNumbers[detectors[0]][:int(maxDataFrameSize)]))
            
            for filenumber in fileNumbers[detectors[0]][:int(maxDataFrameSize)]:

                # For each file we in first detector we check that it exists in the
                # other detectors too. 
                if all( list( os.path.isfile(config['masterDirectory']
                                            +'/temp/'+det+'/'+fileNumbers[detectors[0]][c]
                                            +'_1.pkl') for det in detectors)):
                    _temp_dataSetDict = {}

                    try:
                        for det in detectors:
                                _temp_dataSetDict[det] = DataPod.load(config['masterDirectory']
                                                        +'/temp/'+det+'/'+str(filenumber)+'_1.pkl')
                        for det in detectors:

                            dataSetDict[det].append(_temp_dataSetDict[det])

                    except Exception as e:

                        logger.warning("Empty file detected: "+config['masterDirectory']
                                            +'/temp/'+det+'/'+str(filenumber)+'_1.pkl')

                            
                        
                    lastFileNumber=int(float(filenumber))
                    filesToDelete.append(str(filenumber)+'_1.pkl')
                    
                    
                
                # Othewise we delete it and it can't be used
                else:
                    filesToDelete.append(str(filenumber)+'_1.pkl')
                
                c+=1

            # We use those two for name creation
            first = str(int(float(fileNumbers[detectors[0]][0])))
            last = str(lastFileNumber)
            # print('first,last: ',first,last)
            
            # Saving the instances in dataSets
            for det in detectors:
                _set = DataSet(dataSetDict[det])
                name = first+'-'+last+'_'+str(len(_set))
                _set.save(config['masterDirectory']+'/' + det + '/' + name)

                # # Deleting files that already passed.
                for file in filesToDelete:
                    try:
                        os.system("rm "+config['masterDirectory']+'/temp/'+det+'/'+file)
                    except Exception as e:
                        pass

        logger.debug(f"PROCESSES before efficiency and buffers management: {subprocess.check_output(['pgrep','-c', '-w','-u',config['user_name']]) }")

        # # # Efficiency test buffer management
        #

        if os.path.isdir(config['efficiencies']):
            manage_efficiency_test(config)

        # # # FAR test management 
        #

        manage_FAR_inefernce_files(config)
        
        # Updates latest segment plot
        if datetime.now().hour != lastdatetimehour:
            print('start segments plot')
            segments_plot()
            print('end segments plot')

        # Update FAR plot
        far_plot()
        # # # Update status page
        manage_status_page()  

        # Timing
        logger.info(f"Manager loop time:{time.time()-loopT0}, waiting 5 minutes ... ")
        sys.stdout.flush()

        time.sleep(5*60)
        





def manage_efficiency_test(config = None):

    if config is None:
        with open('config.json') as json_file:
            config = json.load(json_file)

    np.random.seed(config['seed'])
    random.seed(config['seed'])

    # # # Efficiency test buffer management
    
    
    # How many inferences to wait to run the test, roughly how often in seconds

    efficiencyTestBuffer = config["eff_config"]["howOften"]

    # We load all the buffer pods that live in temp
    buffer_set = []
    pod_list = dirlist(config['bufferDirectory']+"/temp")
    # Shuffling the pods so that the ones used are randomly distributed along the time.
    random.shuffle(pod_list)
    
    num_pods = len(pod_list)
    
    # If number of the buffer pods are more than maxDataFrameSize, we 
    # putting them in a dataSet to be used.
    logger.info(f"Buffer size accumulated: {num_pods}/{efficiencyTestBuffer} .Start efficiency test: {num_pods >= efficiencyTestBuffer}")
    # I there are enough buffer pods we can save them to be used in an efficiency test.
    if num_pods >= efficiencyTestBuffer:
        
        for each_pod in pod_list[:efficiencyTestBuffer]:
            pod = DataPod.load(config['bufferDirectory']+"/temp/"+each_pod) 
            buffer_set.append(pod)

        buffer_dataset = DataSet(buffer_set)
        buffer_dataset.save(config['bufferDirectory']+"/BUFFER_SET")

        # Deleting the files in temp
        for each_pod in pod_list:    
            os.remove(config['bufferDirectory']+"/temp/"+each_pod)
        
        issue_efficiency_test()
        logger.info(f"Efficiency test initiated")

def issue_efficiency_test():
    os.system("nohup python -m mly_pipeline.make_eff_estimation --mode condor > efficiencyTest.out &")

def manage_FAR_inefernce_files(config = None):

    if config is None:
        with open('config.json') as json_file:
            config = json.load(json_file)
    # # # FAR test management 
    #

    # The list of all temporary inference files
    if os.path.isdir(config['falseAlarmRates']):
        if config['far_config']['restriction']==None: 
            config['far_config']['restriction']=0.0
        
        restriction=float(config['far_config']['restriction'])
        
        inference_temp_files = dirlist(config['falseAlarmRates'])

        for directory in ['hourly', 'condor', 'FARfile','temp']:
            if directory in inference_temp_files: inference_temp_files.remove(directory)


        # The list of all the files in hourly (grouped by hour/DAG of generation)
        hourly_files = dirlist(config['falseAlarmRates']+"/hourly")

        # The list of all condor dagman directories
        condor_files =  dirlist(config['falseAlarmRates']+"/condor")

        # The group names (start-end gps) in inference_temp_files
        inference_temp_files_groups = list(file.split("_")[0] for file in inference_temp_files)
        inference_temp_files_groups = list(dict.fromkeys(inference_temp_files_groups))

        # The group names inside hourly
        hourly_files_groups = list(file.split("_")[0] for file in hourly_files)

        # The group names of condor directories
        condor_groups= list(file.split("_")[0] for file in condor_files)

        logger.debug(f"HOURLY GROUPS: {len(hourly_files_groups)}")
        logger.debug(f"TEMP GROUPS: {len(inference_temp_files_groups)}")

        # First we will go through the files that already have a group inside hourly
        # We will add them in their corresponding hourly group file and remove them 
        # from inference_temp_files

        for hgroup, hfile_name in zip(hourly_files_groups, hourly_files):

            if hgroup in inference_temp_files_groups:

                with open(config['falseAlarmRates']+"/hourly/"+hfile_name,'rb') as obj:
                    hfile = pickle.load(obj)
                
                hfile_size = int(hfile_name.split("_")[-1][:-4])

                to_remove_from_inference_temp_files = []

                for file_name in inference_temp_files:
                    
                    if hgroup in file_name:

                        with open(config['falseAlarmRates']+"/"+file_name,'rb') as obj:
                            file = pickle.load(obj)

                        file = file[file['score'] >= restriction]

                        hfile = pd.concat([hfile, file], ignore_index=True)
                        
                        # Adding the size of the file to the hourly file
                        hfile_size += int(file_name.split("_")[-1][:-4])

                        #TEMPORARY
                        os.system("rm "+config['falseAlarmRates']+"/"+file_name)
                        to_remove_from_inference_temp_files.append(file_name)

                new_hfile_name = "_".join(hfile_name.split("_")[:-1])+"_"+str(hfile_size)
                with open(config['falseAlarmRates']+"/hourly/"+new_hfile_name+'.pkl', 'wb') as output:
                    pickle.dump(hfile, output, 4)

                os.system("rm "+config['falseAlarmRates']+"/hourly/"+hfile_name)

                for file_name in to_remove_from_inference_temp_files:
                    inference_temp_files.remove(file_name)

        # Now we can create new hourly files for the new groups that appeared
        # First we reset the inference_temp_files. There will be no already existing froup
        # for those files

        # Reseting temp_file_groups 
        inference_temp_files_groups = list(file.split("_")[0] for file in inference_temp_files)
        inference_temp_files_groups = list(dict.fromkeys(inference_temp_files_groups))

        # A check to make sure all files left do not belong in an existing group.
        check = all( [ all([hgroup not in file for file in inference_temp_files]) for hgroup in hourly_files_groups] )
        logger.debug(f"All groups left are new: {check}")


        for tgroup in inference_temp_files_groups:

            to_remove_from_inference_temp_files = []

            for file_name in inference_temp_files:

                if tgroup in file_name:

                    if len(to_remove_from_inference_temp_files)==0:

                        with open(config['falseAlarmRates']+"/"+file_name,'rb') as obj:
                            new_hfile = pickle.load(obj)

                        new_hfile_size = int(file_name.split("_")[-1][:-4])

                        to_remove_from_inference_temp_files.append(file_name)
                    
                    else:

                        with open(config['falseAlarmRates']+"/"+file_name,'rb') as obj:
                            file = pickle.load(obj)

                        new_hfile = pd.concat([new_hfile, file], ignore_index=True)

                        new_hfile_size += int(file_name.split("_")[-1][:-4])

                        to_remove_from_inference_temp_files.append(file_name)

            new_hfile_name = tgroup+"_"+str(new_hfile_size)
            with open(config['falseAlarmRates']+"/hourly/"+new_hfile_name+'.pkl', 'wb') as output:
                pickle.dump(new_hfile, output, 4)

            logger.info(f"Informing / Saving FAR file : {config['falseAlarmRates']}/hourly/{new_hfile_name}.pkl")

            for file_name in to_remove_from_inference_temp_files:
                inference_temp_files.remove(file_name)

                #TEMPORARY
                os.system("rm "+config['falseAlarmRates']+"/"+file_name)

        logger.debug(f"This should be zero, inference_temp_files: {len(inference_temp_files)}")


        hourly_files = dirlist(config['falseAlarmRates']+"/hourly") 

        for i, hfile_name in enumerate(hourly_files):

            if i==0: 
                with open(config['falseAlarmRates']+"/hourly/"+hfile_name,'rb') as obj:
                    farfile = pickle.load(obj)

                farfile_size = int(hfile_name.split("_")[-1][:-4])
            else:
                
                with open(config['falseAlarmRates']+"/hourly/"+hfile_name,'rb') as obj:
                    farfile_part = pickle.load(obj)

                farfile = pd.concat([farfile, farfile_part], ignore_index=True)

                farfile_size += int(hfile_name.split("_")[-1][:-4])

        # Deleting the old FARfile before creating the new one (they have the size of tests on their name).
        farfile_to_delete = None
        old_far_files = dirlist(config['falseAlarmRates']+"/FARfile")
        for file in old_far_files:
            if 'FARfile_interpolation' not in file:
                farfile_to_delete = file
                logger.info(f"FARfile to be replaced: {farfile_to_delete}")
                break 

        if len(hourly_files)!=0:
            with open(config['falseAlarmRates']+"/FARfile/FARfile_"+str(farfile_size)+".pkl", 'wb') as output:
                pickle.dump(farfile, output, 4)

                logger.info(f"New FARfile saved: {config['falseAlarmRates']}/FARfile/FARfile_{farfile_size}.pkl")

            far_interpolation_output = far_interpolation(farfile, farfile_size, inverse = False)
            far_interpolation_output_inverse = far_interpolation(farfile, farfile_size, inverse = True)

            with open(config['falseAlarmRates']+"/FARfile/FARfile_interpolation.pkl", 'wb') as output:
                pickle.dump( far_interpolation_output , output, 4)
            with open(config['falseAlarmRates']+"/FARfile/FARfile_interpolation_reserve.pkl", 'wb') as output:
                pickle.dump( far_interpolation_output , output, 4)
                
            with open(config['falseAlarmRates']+"/FARfile/FARfile_interpolation_inverse.pkl", 'wb') as output:
                pickle.dump(far_interpolation_output_inverse , output, 4)
            with open(config['falseAlarmRates']+"/FARfile/FARfile_interpolation_inverse_reserve.pkl", 'wb') as output:
                pickle.dump(far_interpolation_output_inverse , output, 4)
            
            logger.info(f"New FARfile interpolations saced")

            if (farfile_to_delete is not None 
                    and farfile_to_delete!="FARfile_"+str(farfile_size)+".pkl"):
                os.system("rm "+config['falseAlarmRates']+"/FARfile/"+farfile_to_delete)
            
        # Overwriting config farfile when enough files
        if len(hourly_files)!=0 and (farfile_size) >= 1*365*24*3600 and (config['farfile'] != config['falseAlarmRates']+"/FARfile"):

            with open('config.json') as json_file:
                _config = json.load(json_file)
                _config['farfile'] = config['falseAlarmRates']+"/FARfile"
            with open("config.json", "w") as config_json:
                json.dump(_config, config_json,indent=4)
                config_json.close()

            logger.info(f"FARFILE UPDATED FROM THE TEMPORARY TO THE SEARCH LOCAL!")
            raise SystemError("This is just an interuption to load the new FAR sourse.")


# add here the page function
def manage_status_page():
    os.system("nohup python -m mly_pipeline.create_status_page --mode condor > status_page.out &")

def far_plot():

    farfile_path = config['farfile']


    for file in os.listdir(farfile_path):
        if 'interpolation' not in file:
            main_file = file
            break


    interpolation= farfile_path + '/FARfile_interpolation.pkl'
    inverse_interpolation= farfile_path + '/FARfile_interpolation_inverse.pkl'

    testnums=int(main_file.split('_')[-1][:-4])
    years_of_background = "{:0.2f}".format(testnums/365/24/3600)

    with open(interpolation,'rb') as handle:
        interp= pickle.load(handle)

    with open(inverse_interpolation,'rb') as handle:
        inverse= pickle.load(handle)

    s = np.arange(1e-2,1,1e-4)

    halfday_score = inverse(1/(12*3600))
    month_score = inverse(1/(30*24*3600))
    year_score = inverse(1/(365*24*3600))
    fouryear_score = inverse(1/(4*365*24*3600))

    # print(month_score,year_score,fouryear_score)

    data = 1/(interp(s)*24*3600*365)
    # print(data[-1])
    fig, ax  = plt.subplots()

    ax.axis([s[0], 1, data[-1], data[0]])

    ax.loglog(s,data) # IFAR months
    ax.axvline(x=halfday_score,ls='--',color='black',label ='2/day')#,color=colours[i])
    ax.axvline(x=month_score,ls='--',color='green',label ='1/month')#,color=colours[i])
    ax.axvline(x=year_score,ls='--',color='orange',label = '1/year')#,color=colours[i])
    ax.axvline(x=fouryear_score,ls='--',color='red', label = '1/4years')#,color=colours[i])

    ax.set_title(f"Background score distribution of {years_of_background} years")
    ax.set_ylabel("IFAR (years)")
    ax.set_xlabel("Scores")

    ax.text(halfday_score, 1/1e3, "{:6.3f}".format(halfday_score), color ='black',fontweight ='bold',horizontalalignment = 'right',verticalalignment ='top')
    ax.text(month_score, 1/1e2, "{:6.3f}".format(month_score), color ='green',fontweight ='bold',horizontalalignment = 'right',verticalalignment ='top')
    ax.text(year_score, 1/1e1, "{:6.3f}".format(year_score), color ='orange',fontweight ='bold',horizontalalignment = 'right',verticalalignment ='top')
    ax.text(fouryear_score, 1/1e0, "{:6.3f}".format(fouryear_score), color ='red',fontweight ='bold',horizontalalignment = 'right',verticalalignment ='top')


    ax.legend()

    plt.savefig('far.png')


def segments_plot():


    detectors = config['detectors']

    if 'ONLINE' in config['search_mode']:

        end = int(to_gps(datetime.now().date()))-3600
        start = end-3600*24*3+3600

    elif 'OFFLINE' in config['search_mode'] and isinstance(config['segment_list'],(list,tuple)):

        start = config['segment_list'][0]
        end = config['segment_list'][-1]

    else:

        print('NO VALID SEGMENT INPUT WAS PROVIDED')

    detector_active_flag =  config['active_flags']
    detector_segments =[]

    time0 = time.time()
    for det in detectors:
        individual_detector_segment = query_segments(detector_active_flag[det], start, end)['active']
        detector_segments.append(individual_detector_segment)

    time1 = time.time()-time0

    coincident_segments = SegmentList(detector_segments[0])
    for segment in detector_segments[1:]:
        coincident_segments = coincident_segments & segment

    time2 = time.time()-time1


    files = sorted(os.listdir(config['masterDirectory']+'/'+detectors[0]))
    slist = SegmentList([])

    for file in files:
        try:

            start_gps = int(file.split('-')[0])
            duration = int(file.split('_')[1].split('.')[0])
            slist.append(Segment(start_gps,start_gps+duration + 1 ))
        except Exception as e:

            print(file,':', e)

    slist = slist.coalesce() 

    slist = slist & SegmentList([Segment(start, end)])
    time3 = time.time()-time2

    flag = DataQualityFlag(name=' '
                    , active=slist
                    , known=coincident_segments)

    plot = flag.plot(color='gwpy:ligo-hanford',alpha =0.6)
    #plt.title('Processed Segments the last 3 days')
    plot.savefig('segment_plot.png')


if __name__ == "__main__":
        
    main(config)
    





#!/usr/bin/env python

import os
import time ,json, argparse, subprocess

import numpy as np
import pickle
import matplotlib.pyplot as plt

from gwpy.time import from_gps
from pycondor import Job, Dagman
from mly.datatools import DataSet
from mly.validators import Validator
from mly.tools import *


from .search_functions import far, stackVirgo

with open('config.json') as json_file:
    config = json.load(json_file)

config = stackVirgo(config)

np.random.seed(config['seed'])

#os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

which_python = str(subprocess.check_output(['which','python']))[2:-3]
print(which_python)

# # # Managing arguments
# 
# List of arguments to pass:
arguments = ['mode','injectionPath','injectionSNR','injectionHRSS','group_directory']

#Construct argument parser:
parser = argparse.ArgumentParser()

parser.add_argument('--mode')
parser.add_argument('--injectionPath')
parser.add_argument('--injectionSNR' , default = None)
parser.add_argument('--injectionHRSS', default = None)
parser.add_argument('--group_directory', default = None)

# Pass arguments
args = parser.parse_args()

# Store arguments in dictionary:
kwdict = {}
for argument in arguments:
    kwdict[argument] = getattr(args, argument)


def moving_average(x, w):
    ma=[]
    for i in range(len(x)):
        if i<int(w/2)+1:
            ma.append(np.mean(x[:i+1]))
        elif i>=len(x)-(w/2+1):
            ma.append(np.mean(x[i:]))
        else:
            ma.append(np.mean(x[i-int(w/2):i+int(w/2)]))

    return np.array(ma)


def main():
    
    mode = kwdict['mode']
    
    if mode=='test':

        size = int(float(config["eff_config"]["testSize"]))
        injection_path = kwdict['injectionPath']
        noiseSource = config["bufferDirectory"]+"/BUFFER_SET.pkl"
        injectionSNR = kwdict["injectionSNR"]
        injectionHRSS = kwdict["injectionHRSS"]
        group_directory = kwdict["group_directory"]


        print("INJSNR",injectionSNR, type(injectionSNR))
        print("INJHRSS",injectionHRSS, type(injectionHRSS))
        if injectionSNR != None:

            injectionSNR = np.array(list(float(_) for _ in injectionSNR.split(',')))
            injectionSNR = np.arange(injectionSNR[0], injectionSNR[1], injectionSNR[2]).tolist()

        if injectionHRSS != None:
            injectionHRSS = np.array(list(float(_) for _ in injectionHRSS.split(',')))
            injectionHRSS = np.arange(injectionHRSS[0], injectionHRSS[1], injectionHRSS[2]).tolist()


        print("INJSNR",injectionSNR)
        print("INJHRSS",injectionHRSS)
        noiseSource = DataSet.load(noiseSource)
        gps_time_start = int(noiseSource[0].gps[0])
        gps_time_end = int(noiseSource[-1].gps[0])

        models = [[config['model1_path'],config['model2_path']]
                  ,[["strain"], ["strain", "correlation"]]]

        # print(models)
        # print(injectionSNR)
        # print(injectionHRSS)
        print('injection_path: ',injection_path)
        t0=time.time()
        print("savefile",group_directory+f"{injection_path.split('/')[-1]}_{gps_time_start}-{gps_time_end}")

        if 'V' not in config['detectors']:
            efficiencies = Validator.accuracy(models=models, 
                                        duration=1, 
                                        size=size,
                                        fs=1024,
                                        detectors=config['detectors'], 
                                        labels={"type": "signal"},
                                        backgroundType="real",
                                        injection_source=injection_path,
                                        injectionSNR = injectionSNR,
                                        noiseSourceFile = noiseSource,
                                        windowSize = 16, 
                                        plugins=["correlation_30"], 
                                        mapping=2 * [{"noise":[ 1, 0], "signal":[0, 1]}],
                                        name=f"{injection_path.split('/')[-1]}_{gps_time_start}-{gps_time_end}",
                                        savePath = group_directory,
                                        injectionHRSS = injectionHRSS, #list(np.arange(0, 100) * 1e-23)
                                        stackDetectorDict = config['stackDetectorDict']
            )   


        else:
            efficiencies = Validator.accuracy(models=models, 
                                duration=1, 
                                size=size,
                                fs=1024,
                                detectors=config['detectors'], 
                                labels={"type": "signal"},
                                backgroundType="real",
                                injection_source=injection_path,
                                injectionSNR = injectionSNR,
                                noiseSourceFile = noiseSource,
                                windowSize = 16, 
                                plugins=["correlation_30"], 
                                mapping=2 * [{"noise":[ 1, 0], "signal":[0, 1]}],
                                name=f"{injection_path.split('/')[-1]}_{gps_time_start}-{gps_time_end}",
                                savePath = group_directory,
                                injectionHRSS = injectionHRSS
                                )#list(np.arange(0, 100) * 1e-23))]]

        print(time.time()-t0)
        #return efficiencies
    
    elif mode=='plot':
        
        farfile_inverse_interpolation = config["farfile"]+"/FARfile_interpolation_inverse.pkl"
        group_directory = kwdict["group_directory"]

        thresholds = {#'10year': far(farfile_inverse_interpolation, 1/(10*365*24*3600) , inverse = True)
                      '4years': far(farfile_inverse_interpolation, 1/(4*365*24*3600) , inverse = True)
                      ,'1year': far(farfile_inverse_interpolation, 1/(1*365*24*3600) , inverse = True)
                      ,'12hours': far(farfile_inverse_interpolation, 1/(12*3600) , inverse = True)

                     }

        print(thresholds)
        
        tkeys=list(thresholds.keys())    
        
        tests= dirlist(group_directory)

        tests.remove('condor')
        
        buffer_start_time = from_gps(int(tests[0].split("_")[-1][:-4].split('-')[0]))
        buffer_extention = f"{(config['eff_config']['howOften']/3600):.1f}"

        colours=['#377eb8', '#ff7f00', '#4daf4a',
                  '#f781bf', '#a65628', '#984ea3',
                  '#999999', '#e41a1c', '#dede00']

        fig,ax=plt.subplots(2,3,figsize=(15,10))

        ACC50={}

        for t in range(len(tkeys)):
            ACC50[tkeys[t]]={}

            for l in range(len(tests)):
                label = tests[l].split('_')[0]

                with open(group_directory+tests[l],'rb') as handle: data= pickle.load(handle)

                acc=[]
                
                if 'hrss' in data.keys():
                    looper = data['hrss']
                    xlabel = 'hrss'
                    
                    for hrss in range(len(looper)):
                        count=0
                        for sc in np.array(data['scores1'][hrss])*np.array(data['scores2'][hrss]):
                            if sc>=thresholds[tkeys[t]]: count+=1
                        acc.append(100*count/len(np.array(data['scores1'][hrss])*np.array(data['scores2'][hrss])))
                    
                    #ax[0,t].plot(looper,moving_average(acc,10),label=label,color=colours[l])
                    ax[0,t].plot(looper,acc,label=label,color=colours[l])

                    mvave=moving_average(acc,10)
                    for snr in range(1,len(looper)):

                        if mvave[snr]>50 and mvave[hrss-1]<=50:
                            ACC50[tkeys[t]][label]=looper[snr]


                    ax[0,t].set_title(f"IFAR {tkeys[t]}")

                    ax[0,t].set_xlabel(xlabel)
                    ax[0,t].set_ylim(0,100)
                    ax[0,t].grid(True,which='both')



                elif 'snrs' in data.keys():
                    looper = data['snrs']
                    xlabel = 'SNR'
                    
                    
                    for snr in range(len(looper)):
                        count=0
                        for sc in np.array(data['scores1'][snr])*np.array(data['scores2'][snr]):
                            if sc>=thresholds[tkeys[t]]: count+=1
                        acc.append(100*count/len(np.array(data['scores1'][snr])*np.array(data['scores2'][snr])))
                        
                    #ax[1,t].plot(looper,moving_average(acc,10),label=label,color=colours[l])
                    ax[1,t].plot(looper,acc,label=label,color=colours[l])

                    mvave=moving_average(acc,10)
                    for snr in range(1,len(looper)):

                        if mvave[snr]>50 and mvave[snr-1]<=50:
                            ACC50[tkeys[t]][label]=looper[snr]

                    ax[1,t].set_xlabel(xlabel)
                    ax[1,t].set_ylim(0,100)
                    ax[1,t].grid(True,which='both')

                
        ax[0,0].set_ylabel("% of signals detected")
        ax[1,0].set_ylabel("% of signals detected")

        ax[0,2].legend()
        ax[1,2].legend()
        fig.suptitle(f"Efficiencies at {buffer_start_time} + {buffer_extention}h", fontsize=25)

        fig.savefig('./Efficiencies.png')

        os.system('cp ./Efficiencies.png '
                    +group_directory+'/Efficiencies.png')

            
    elif mode=='condor':
        
        
        noiseSource = config["bufferDirectory"]+"/BUFFER_SET.pkl"
        noiseSource = DataSet.load(noiseSource)
        gps_time_start = int(noiseSource[0].gps[0])
        gps_time_end = int(noiseSource[-1].gps[0])
        group_directory = f"{config['efficiencies']}/history/{gps_time_start}-{gps_time_end}/"

        os.mkdir(group_directory)


        accounting_group_user=config['accounting_group_user']
        accounting_group=config['accounting_group']

        injSourceDir = config["eff_config"]["injectionDirectoryPath"]
        
        error = group_directory+'condor/error'
        output = group_directory+'condor/output'
        log = group_directory+'condor/log'
        submit = group_directory+'condor/submit'

        dagman = Dagman(name='efficiencyTestDagman',submit=submit)


        injection_paths_HRSS =  config["eff_config"]["injectionsWithHRSS"]
        
        injection_paths_SNR = config["eff_config"]["injectionsWithSNR"]

        
        job_list=[]

        for injection_path in injection_paths_HRSS:
            print(injSourceDir+injection_path)

            jobname = injection_path.split('/')[-1]
            print(injSourceDir+injection_path)
            thearguments = ( "-m mly_pipeline.make_eff_estimation"
                             +" --mode=test" 
                             +" --injectionPath="+injSourceDir+injection_path
                             +" --injectionHRSS="+config['eff_config']['injectionHRSS']
                             +" --group_directory="+group_directory)



            job = Job(name = jobname
                       ,executable = which_python
                       ,arguments = thearguments
                       ,submit=submit
                       ,error=error
                       ,output=output
                       ,log=log
                       ,getenv=True
                       ,dag=dagman
                       ,extra_lines=["accounting_group_user="+accounting_group_user
                                    ,"accounting_group="+accounting_group
                                    ,"request_memory = 6144M"
                                    ,"request_disk            = 64M"])

            job_list.append(job)
            
            
            
            
            
        for injection_path in injection_paths_SNR:

            jobname = injection_path.split('/')[-1]
            print(injSourceDir+injection_path)
            thearguments = ( "-m mly_pipeline.make_eff_estimation"
                             +" --mode=test" 
                             +" --injectionPath="+injSourceDir+injection_path
                             +" --injectionSNR="+config['eff_config']['injectionSNR']
                             +" --group_directory="+group_directory)





            job = Job(name = jobname
                       ,executable = which_python
                       ,arguments = thearguments
                       ,submit=submit
                       ,error=error
                       ,output=output
                       ,log=log
                       ,getenv=True
                       ,dag=dagman
                       ,extra_lines=["accounting_group_user="+accounting_group_user
                                    ,"accounting_group="+accounting_group
                                    ,"request_memory = 6144M"
                                    ,"request_disk            = 64M"])

            job_list.append(job)


        plot_job = Job(name = 'plotJob'
                   ,executable = which_python
                   ,arguments = ( " -m mly_pipeline.make_eff_estimation"
                                 +" --mode=plot"
                                 +" --group_director="+group_directory)
                   ,submit=submit
                   ,error=error
                   ,output=output
                   ,log=log
                   ,getenv=True
                   ,dag=dagman
                   ,extra_lines=["accounting_group_user="+accounting_group_user
                                ,"accounting_group="+accounting_group
                                ,"request_disk            = 64M"])

        plot_job.add_parents(job_list)


        dagman.build_submit()

if __name__ == "__main__":
    main()



#! /usr/bin/env python3

import os, sys, json, argparse, subprocess

def _parser():

    arguments = ["search_mode"
                 ,"path"
                 ,"output_directory"
                 ,"trigger_directory"
                 ,"masterDirectory"
                 ,"bufferDirectory"
                 ,"falseAlarmRates"
                 ,"efficiencies"
                 ,"detectors"
                 ,"reset"
                 ,"nofar"]

    
    #Construct argument parser:
    parser = argparse.ArgumentParser()

    parser.add_argument('-s', '--search_mode', type=str
                        , default = "DEFAULT_OFFLINE"
                        , help = "For automatically setting the pipeline for different uses"
                        , required = False)

    parser.add_argument('-p', '--path', type=str
                        , default = "./mlyPipelineOrigin"
                        , help = "The path and name of the"
                        +" directory to be used by mlyPipeline."
                        , required = False)

    parser.add_argument('-o', '--output_directory', type=str
                        , default = 'output_directory'
                        , help = "The name of the"
                        +" directory to be used by mlyPipeline for the storage"
                        +" of the results from every inference."
                        , required = False)

    parser.add_argument('-t', '--trigger_directory', type=str
                        , default = 'trigger_directory'
                        , help = "The name of the"
                        +" directory to be used by mlyPipeline for the storage"
                        +" of the triggers."
                        , required = False)

    parser.add_argument('-m', '--masterDirectory', type=str
                        , default = 'masterDirectory'
                        , help = "The name of the"
                        +" directory to be used by mlyPipeline for the storage"
                        +" of the plots from the triggers."
                        , required = False)

    parser.add_argument('-b', '--bufferDirectory', type=str
                        , default = 'bufferDirectory'
                        , help = "The name of the"
                        +" directory to be used by mlyPipeline for the storage"
                        +" of the data around every instance (buffer) to be used"
                        +" for the efficiecy tests"
                        , required = False)

    parser.add_argument('-f', '--falseAlarmRates', type=str
                        , default = 'falseAlarmRates'
                        , help = "The name of the"
                        +" directory to be used by mlyPipeline for the storage"
                        +" of the false alarm rate data."
                        , required = False)

    parser.add_argument('-e', '--efficiencies', type=str
                        , default = 'efficiencies'
                        , help = "The name of the"
                        +" directory to be used by mlyPipeline for the storage"
                        +" of the false alarm rate data."
                        , required = False)

    parser.add_argument('-d', '--detectors', type=str
                        , default = 'HLV'
                        , help = "The initials of the detectors used"
                        , required = False)
    
    parser.add_argument('-R', '--reset'
                        , action = 'store_true'
                        , default = False
                        , help = "If passed it will ovewright the target path"
                        + " with a clean new filesystem."
                        , required = False)

    parser.add_argument('-nofar', '--nofar'
                        , action = 'store_true'
                        , default = False
                        , help = "If passed it will exclude the FAR scripts used for background estimation. This is only for OFFLINE searches"
                        , required = False)




    
    # Pass arguments:
    args = parser.parse_args()
    
    # Store arguments in dictionary:
    kwargs = {}
    for argument in arguments:
        kwargs[argument] = getattr(args, argument)

    print('nofar: ',kwargs['nofar'])

    return kwargs
   
def connect_path_relatively_to(path_to_change,path_to_connect):
    common_part = []
    for dir in path_to_change.split('/'):
        if dir in path_to_connect.split('/'):
            common_part.append(dir)
        else:
            break

    common_part = '/'.join(common_part)

    new_path_to_change = path_to_change.split(common_part)
    
    return '..'+new_path_to_change[-1]

def createRunAllScript(**kwargs):
    
    """ Creating a script called runall.sh that withing the search director will
    run and manage all the scripts needed for the search.
    """
    search_directory = kwargs['path'].split('/')[-1]
    with open(kwargs['path'] + '/' + 'runall.sh','w') as f:
        
        # Sourcing mlyPipeline module
        #f.write("source=" + kwargs['mlyPipelineSource'] + "\n")
        
        # Things to write only if OFFLINE mode is active 
        # (OFFLINE is not complete)
        f.write("echo $(date +'%Y-%m-%d %T') SEARCH HAS STARTED > monitor.log" + "\n")
        f.write("echo $(date +'%Y-%m-%d %T') Version of mly: $(python -m pip show mly 2>/dev/null | grep Version | awk '{print $2}') >> monitor.log" + "\n")
        f.write("echo $(date +'%Y-%m-%d %T') Version of mly-pipeline: $(python -m pip show mly_pipeline 2>/dev/null | grep Version | awk '{print $2}') >> monitor.log" + "\n")     

        if "ONLINE" in kwargs['search_mode']:
            
            # Loop that checks the state of the jobs and doesn't stop
            f.write("while true" + "\n")
            f.write("do" + "\n")
            
            # Collection of the process ids of the scripts
            f.write("   processID=()"+ "\n")
            # Used to sinchronise the independent scripts of the search
            f.write("   unixtime=$(date +%s)" + "\n")
            # The number of paralel schripts to run 
            # !Note: If you change this from the config, you will have to 
            #        edit in the runall.sh script by hand.
            f.write("   STEP="+str(kwargs['parallel_scripts'])+ "\n\n")
            
            # Lopp to run the scripts for the search
            f.write("   for INITIAL in `seq 0 $(($STEP-1))`" + "\n")
            f.write("   do" + "\n\n")
            f.write("       nohup python -m mly_pipeline.search"
                            +" --splitter [$STEP,$INITIAL]"
                            +" --time_reference $unixtime" 
                            +" &> search_step_$INITIAL.out & processID+=($(echo $!)) ; echo $!>>.jobids.txt" + "\n\n")
            f.write("   echo $(date +'%Y-%m-%d %T') Search file with initial $INTIAL started >> monitor.log" + "\n")     
            f.write("   done" + "\n\n\n")
        
            # Script that runs the manager script
            f.write("   nohup python -m mly_pipeline.manager &> manager.out & processID+=($(echo $!)) ; echo $!>>.jobids.txt" + "\n")
            f.write("   echo $(date +'%Y-%m-%d %T') Manager script started >> monitor.log" + "\n")     

            # If the search runs as expected by default, it will also run scripts
            # for generating timelags and running FAR tests.
            if kwargs['far_config']['far_enabled']:
                f.write("   nohup python -m mly_pipeline.continuous_FAR --mode generation &> continuous_FAR_generation.out & generationID+=($(echo $!)) ; echo $!>>.jobids.txt" + "\n")
                f.write("   echo $(date +'%Y-%m-%d %T') Continuous FAR generation script started >> monitor.log" + "\n")     

                f.write("   nohup python -m mly_pipeline.continuous_FAR --mode inference &> continuous_FAR_inference.out & inferenceID=($(echo $!)) ; echo $!>>.jobids.txt" + "\n")
                f.write("   echo $(date +'%Y-%m-%d %T') Continuous FAR inference script started >> monitor.log" + "\n")     
            # # # Maintenance of jobs 
            
            # If total_failure becomes 1, the script runs everything from scratch
            f.write("    total_failure=0" + "\n")
            
            # Sometimes the FAR script might stal indefinetly for unknow reasons
            # related to triton package, so if it does we restart it.
            if kwargs['far_config']['far_enabled']:

                f.write("    stalling_points=0" + "\n")
            
            # We add the number of timelag files present to see if our rate of 
            # doing FAR inference is efficient. Check the monitor.txt present in
            # the search directory, if there are too many zeros we can increase
            # our timelag production - "lags", always along with "batches". If
            # there are no zeros, we need to decrease the "lags" or inference 
            # does not work properly.
                    
            # Loop that checks if the main (not inference) scripts are running
            # every minute.
            f.write("    while [ $total_failure -ne 1 ]" + "\n")
            f.write("    do" + "\n")
            f.write("        sleep 60" + "\n\n")
                    
            # Updating the number of lag files for monitoring
            if kwargs['far_config']['far_enabled']:
                f.write("        echo $(date +'%Y-%m-%d %T') Continuous FAR status: $(ls falseAlarmRates/temp | wc -l) files in inference queue >> monitor.log" + "\n")     

            # Checking all scripts (not inference), if one fails total_failure=1
            f.write("        for pid in \"${processID[@]}\"" + "\n")
            f.write("        do" + "\n")
                    
            f.write("            if ! ps $pid > /dev/null" + "\n")
            f.write("            then" + "\n")
            f.write("                total_failure=1" + "\n")
            f.write("                echo -e \"$(date +'%Y-%m-%d %T') \\033[1mMain search script with pid $pid exited unexpectedly. Restarting search sequence ... \\033[0m \" >> monitor.log" + "\n")     
            f.write("                break" + "\n")
            f.write("            fi" + "\n") 
            f.write("        done" + "\n")

            # Specifically for inference script we check seperatelly, if it
            # doesn't look properly we rerun only the inference script. 
            if kwargs['far_config']['far_enabled']:

            
                # Checking if the number of inference files stays 0 for a long time.
                # If lag files increase without any new inference files, that means
                # stalling, and we add a stalling point. If not we reset it to 0.
                f.write("        if [ $(ls falseAlarmRates | wc -l) -eq 4 ] && [ $(ls falseAlarmRates/temp | wc -l) -ne 0 ]" + "\n")
                f.write("        then" + "\n")
                f.write("            echo $(date +'%Y-%m-%d %T') Continuous FAR status: Stalling points increase to $stalling_points >> monitor.log" + "\n")     
                f.write("            ((stalling_points=stalling_points+1))" + "\n")
                f.write("        else" + "\n")
                f.write("            stalling_points=0" + "\n")
                f.write("        fi" + "\n")

                # In case inference script gets killed or we have at least 5 minutes
                # of stalling, we restart inference script. Although, we keep the 
                # corresponding unix time it got killed in the name of the  output
                # file.
                f.write("        if ! ps $inferenceID > /dev/null || [ $stalling_points -ge 5 ]"+ "\n")
                f.write("        then" + "\n")
                f.write("            singularity instance stop --all" + "\n")
                f.write("            nohup python -m mly_pipeline.continuous_FAR --mode inference >"
                                    +" continuous_FAR_inference_$(date +%s).out 2>&1"
                                    +" & inferenceID=($(echo $!)) ; echo $!>>.jobids.txt" + "\n")
                f.write("            echo $(date +'%Y-%m-%d %T') Continuous FAR status: Continuous FAR inference script restarted - Stalling points $stalling_points/5 >> monitor.log" + "\n")     
                f.write("            stalling_points=0" + "\n")
                f.write("        fi" + "\n")

                f.write("        if ! ps $generationID > /dev/null"+ "\n")
                f.write("        then" + "\n")
                f.write("            nohup python -m mly_pipeline.continuous_FAR --mode generation &> "
                                    +"continuous_FAR_generation.out & generationID=($(echo $!)) ; echo $!>>.jobids.txt"+ "\n")
                f.write("            echo $(date +'%Y-%m-%d %T') Continuous FAR status: Continuous FAR generation script restarted due to unexpected exit >> monitor.log" + "\n")     
                f.write("        fi" + "\n\n")  
                    
            f.write("   done" + "\n")
                    
            # If the previous loop exits, it means there was a total_failure=1
            # So we will have to rerun everything. But first we kill the python
            # scripts (inference included).
            # !Note: We are not remove any condor jobs from the lag generation
            #        script because they do not interfere with the scripts.
            f.write("   for pid in \"${processID[@]}\"" + "\n")
            f.write("   do" + "\n")
            f.write("       kill $pid" + "\n")
            f.write("   done" + "\n")
            
            if kwargs['far_config']['far_enabled']:
                
                f.write("   kill $inferenceID" + "\n")
                f.write("   singularity instance stop --all" + "\n")

            f.write("    echo $(date +'%Y-%m-%d %T') SEARCH HAS RESTARTED  >> monitor.log" + "\n")     
            f.write("done" + "\n")
        
        # Currently the offline script is very simple (under developement)
        else:
            
            if kwargs['far_config']['far_enabled']:

                f.write("echo $(date +'%Y-%m-%d %T') Setting file system - Creating condor jobs >> monitor.log" + "\n")
                f.write("python -m mly_pipeline.offline_search --mode set_file_system" + "\n\n")
                f.write("echo $(date +'%Y-%m-%d %T') Setting file system - Condor jobs submited >> monitor.log" + "\n\n")

                f.write("dagman_out_file=$(find masterDirectory/condor/submit -type f -name '*dagman.out' -exec ls -t {} + | head -n 1)" + "\n")
                f.write(f"total_jobs_number=$(grep -oP 'Dag contains \K\d+' $dagman_out_file)" + "\n")

                f.write("while ! grep -q 'EXITING WITH STATUS' \"$dagman_out_file\"; do" + "\n")
                f.write(f"   sleep 300" + "\n")
                f.write(f"   count=$(grep -c 'completed successfully' $dagman_out_file)" + "\n") 
                f.write("    echo $(date +'%Y-%m-%d %T') Setting file system - Completed jobs: $count/$total_jobs_number>> monitor.log" + "\n")
                f.write(f"done" + "\n")
                f.write(f"exiting_status=$(grep -oP 'EXITING WITH STATUS \K\d+' \"$dagman_out_file\")" + "\n")

                f.write(f"created_background_files=$(python -c 'from mly_pipeline.initialization import get_file_counts_after_createFileSystem; get_file_counts_after_createFileSystem()') "+ "\n")
                
                f.write("echo $(date +'%Y-%m-%d %T') Setting file system - $created_background_files/$total_jobs_number, with EXITING WITH STATUS $exiting_status>> monitor.log" + "\n")
                f.write("echo $(date +'%Y-%m-%d %T') Setting file system - Completed >> monitor.log" + "\n")
                
                f.write("nohup python -m mly_pipeline.continuous_FAR --mode generation &> continuous_FAR_generation.out & generationID=($(echo $!)) ; echo $!>>.jobids.txt" + "\n")
                f.write("nohup python -m mly_pipeline.continuous_FAR --mode inference &> continuous_FAR_inference.out & inferenceID=($(echo $!)) ; echo $!>>.jobids.txt" + "\n")
                f.write("echo $(date +'%Y-%m-%d %T') Background estimation initiated >> monitor.log" + "\n")

                # Checking if the number of inference files stays 0 for a long time.
                # If lag files increase without any new inference files, that means
                # stalling, and we add a stalling point. If not we reset it to 0.
                
                f.write("stalling_points=0" + "\n")
                f.write("premature_ending_points=0" + "\n")

                f.write("# The batches of generation times the dagmans" + "\n")
                f.write("background_scripts=$(ls masterDirectory/*.py | wc -l)" + "\n")


                f.write("# While the number of far files generated is not equal to background_scripts" + "\n")
                f.write("while [ $(ls falseAlarmRates/hourly | wc -l) -lt $background_scripts ] || [ $(ls falseAlarmRates/temp | wc -l) -ne 0 ] || (condor_q -af JobBatchName | grep -q \""+search_directory+"\") " + "\n")
                f.write("do" + "\n")
                

                f.write("    echo $(date +'%Y-%m-%d %T') Background estimation - completion: $(ls falseAlarmRates/hourly | wc -l) / $background_scripts >> monitor.log" + "\n")
                f.write("    sleep 300" + "\n\n")

                f.write("    if [ $(ls falseAlarmRates | wc -l) -eq 4 ] && [ $(ls falseAlarmRates/temp | wc -l) -ne 0 ]" + "\n")
                f.write("    then" + "\n")
                f.write("        echo stalling points increase to $stalling_points" + "\n")   
                f.write("        ((stalling_points=stalling_points+1))" + "\n")
                f.write("        echo $(date +'%Y-%m-%d %T') Background estimation - completion: $(ls falseAlarmRates/hourly | wc -l) / $background_scripts, inference stalling points increased to $stalling_points/5 >> monitor.log" + "\n")
                f.write("    else" + "\n")
                f.write("        stalling_points=0" + "\n")
                f.write("    fi" + "\n\n")

                # In case inference script gets killed or we have at least 5 minutes
                # of stalling, we restart inference script. Although, we keep the 
                # corresponding unix time it got killed in the name of the  output
                # file.

                f.write("    if [ $stalling_points -ge 5 ]"+ "\n")
                f.write("    then"+ "\n")
                f.write("        kill $inferenceID"+ "\n")
                f.write("        singularity instance stop --all"+ "\n")
                f.write("    fi"+ "\n\n")

                f.write("    if ! ps $inferenceID > /dev/null"+ "\n")
                f.write("    then" + "\n")
                f.write("        nohup python -m mly_pipeline.continuous_FAR --mode inference &> continuous_FAR_inference.out & inferenceID=($(echo $!)) ; echo $!>>.jobids.txt" + "\n")
                f.write("        stalling_points=0" + "\n")
                f.write("    fi" + "\n\n")       

                f.write("    if ! ps $generationID > /dev/null"+ "\n")
                f.write("    then" + "\n")
                f.write("        nohup python -m mly_pipeline.continuous_FAR --mode generation &> "
                                +"continuous_FAR_generation.out & generationID=($(echo $!)) ; echo $!>>.jobids.txt"+ "\n")
                f.write("    fi" + "\n\n")                    
                        
                f.write("    if ! (condor_q -af JobBatchName | grep -q \""+search_directory+"\")  "
                            +"&& [ $(($(ls falseAlarmRates | wc -l) - 4)) -eq 0 ] "
                            +"&& [ $(ls falseAlarmRates/temp | wc -l) -eq 0 ]" + "\n")
                f.write("    then" + "\n")
                f.write("        ((premature_ending_points=premature_ending_points+1))" + "\n")
                f.write("        echo $(date +'%Y-%m-%d %T') Background estimation - completion: $(ls falseAlarmRates/hourly | wc -l) / $background_scripts, premature ending points increased: $premature_ending_points/3 >> monitor.log" + "\n")
                f.write("    fi" + "\n\n")

                f.write("    if [ $premature_ending_points -ge 3 ]" + "\n")
                f.write("    then " + "\n")
                f.write("        echo $(date +'%Y-%m-%d %T') Background estimation - PREMATURE COMPLETION: $(ls falseAlarmRates/hourly | wc -l) / $background_scripts  >> monitor.log" + "\n")

                f.write("        break" + "\n")
                f.write("    fi" + "\n\n")

                f.write("    python -c \"from mly_pipeline.manager import manage_FAR_inefernce_files; manage_FAR_inefernce_files()\" "+ "\n")


                f.write("done" + "\n")

                f.write("failed_groups=$(comm -23 <(ls falseAlarmRates/condor | awk -F'_' '{print $1}' | sort) <(ls falseAlarmRates/hourly | awk -F'_' '{print $1}' | sort))"+ "\n")
                f.write("failed_groups_n=$(echo $failed_groups | wc -w)"+ "\n")

                f.write("echo $(date +'%Y-%m-%d %T') Background estimation - COMPLETED: $(ls falseAlarmRates/hourly | wc -l)/$(( background_scripts - failed_groups_n )) with $failed_groups_n FAILED groups>> monitor.log" + "\n")
                f.write("[ -n \"$failed_groups\" ] && echo $(date +'%Y-%m-%d %T') Background estimation - The following groups did not generate timelags: $failed_groups"+ "\n")

                f.write("kill $inferenceID $generationID" + "\n")
                f.write("singularity instance stop --all" + "\n")

                # Making the FAR plot for the status page
                f.write("echo $(date +'%Y-%m-%d %T') Generating FAR and Segment plots >> monitor.log" + "\n")
                f.write("python -c \"from mly_pipeline.manager import far_plot; far_plot();\" "+ "\n")

                f.write("python -c \"from mly_pipeline.manager import manage_FAR_inefernce_files; manage_FAR_inefernce_files()\" "+ "\n")
        
            f.write("python -c \"from mly_pipeline.manager import segments_plot; segments_plot();\" "+ "\n")

            # If FAR is already calculated we only need this to run
            f.write("echo $(date +'%Y-%m-%d %T') Offline search - Initialised >> monitor.log" + "\n")
            f.write("python -m mly_pipeline.offline_search \n\n")
            f.write("search_jobs=$(awk -F 'Offline search - Number of jobs submited: ' '/Offline search - Number of jobs submited:/ {jobs=$2} END {print jobs}' monitor.log)" + "\n\n")
            f.write("while [ $(ls output_directory | wc -l) -lt $search_jobs ] " + "\n")
            f.write("do" + "\n")
            f.write("    echo $(date +'%Y-%m-%d %T') Offline search - Search jobs completed: $(ls output_directory | wc -l)/$search_jobs >> monitor.log" + "\n")
            f.write("    sleep 300" + "\n")
            f.write("    if ! condor_q -af JobBatchName | grep -q \""+search_directory+"_offline_search\"" + "\n")
            f.write("    then " + "\n")
            f.write("        echo $(date +'%Y-%m-%d %T') Offline search - WARNING: Condor jobs have finished before all the search output files have been produced - $(ls output_directory | wc -l)/$search_jobs"+ "\n")
            f.write("        break" + "\n\n")            
            f.write("    fi" + "\n\n")            
            f.write("done" + "\n")            
            f.write("echo $(date +'%Y-%m-%d %T') Offline search - Search jobs completed: $(ls output_directory | wc -l)/$search_jobs >> monitor.log" + "\n")

            f.write("if [ $(ls output_directory | wc -l) -ne $search_jobs ]" + "\n")
            f.write("    then " + "\n")
            f.write("    dagman_out_file=$(find condor/submit -type f -name '*dagman.out' -exec ls -t {} + | head -n 1)" + "\n")
            f.write("    error_line=$(grep -n \"ERROR: the following job(s) failed:\" \"$dagman_file\" | cut -d \":\" -f 1)" + "\n")
            f.write("    lines=$(awk -v error_line=\"$error_line\" 'NR > error_line && /Job Submit File:/ { sub(/.*Job Submit File:/, \"\"); print }' \"$dagman_file\")" + "\n")
            f.write("    echo $(date +'%Y-%m-%d %T') The following scripts failed:" + "\n")
            f.write("    while IFS= read -r submit_line; do error_file=$(echo \"$submit_line\" | sed 's/submit/error/g' | tr -d '[:space:]'); echo \"    $error_file - ERROR: $(tail -n 1 \"$error_file\")\" >> monitor.log; done <<< \"$lines\"" + "\n")
            f.write("fi" + "\n\n")            

            f.write("echo $(date +'%Y-%m-%d %T') Creating status page >> monitor.log" + "\n")
            f.write("python -c \"from mly_pipeline.manager import segments_plot; segments_plot();\" "+ "\n")

            f.write("mly-pipeline-statuspage" + "\n")
            f.write("echo $(date +'%Y-%m-%d %T') SEARCH HAS FINISHED >> monitor.log" + "\n")
            f.write("echo SEARCH PAGE:  https://ldas-jobs.ligo.caltech.edu/~$USER/$(jq -r '.path' config.json | awk -F'/' '{print $NF}')/status.html >> monitor.log" + "\n")



def main():
    
    kwargs = _parser()

    if kwargs['path'][-1]=="/" : kwargs['path'] = kwargs['path'][:-1]
    
    
    if os.path.isdir(kwargs['path']) and kwargs['reset'] == True:
                
        
        print(kwargs['path']," is going to be reset")
        print("Are you sure you want to delete the contents of that directory?")
        answer = input()
        
        if answer in ['yes','y','YES','Yes','Y']:
            
            os.system("rm -r "+kwargs['path'])
            os.system("mkdir "+kwargs['path'])
            
        else:
            
            raise ValueError("Initialization stoped from user")

    elif os.path.isdir(kwargs['path']) and kwargs['reset'] == False:
        
        raise ValueError("The directory you try to create already exist"
                         +". Run the same command but with `-R or --reset` to "
                         +"ovewright the directory.")
    
    elif not os.path.isfile(kwargs['path']):
        
        os.system("mkdir "+kwargs['path'])
        
    else:
        raise ValueError("Wrong input for file", os.path.isfile(kwargs['path'])
                         , kwargs['reset'] == True)
        
    print(f"Search mode selected: {kwargs['search_mode']}")

    frames_dict = dict(   DEFAULT_ONLINE = {"H":"" ,"L":"" ,"V":""}

                        ,DEFAULT_OFFLINE = {"H":"" ,"L":"" ,"V":""}

                        , O4_ONLINE = {"H":"/dev/shm/kafka/H1/"
                                      ,"L":"/dev/shm/kafka/L1/"
                                      ,"V":"/dev/shm/kafka/V1/"}

                        , O3MDC_ONLINE = {"H":"/dev/shm/kafka/H1_O3ReplayMDC/"
                                  ,"L":"/dev/shm/kafka/L1_O3ReplayMDC/"
                                  ,"V":"/dev/shm/kafka/V1_O3ReplayMDC/"}

                        , O3REPLAY_ONLINE = {"H":"/dev/shm/kafka/H1_O3ReplayMDC/"
                                     ,"L":"/dev/shm/kafka/L1_O3ReplayMDC/"
                                     ,"V":"/dev/shm/kafka/V1_O3ReplayMDC/"}

                        , BENCHMARK_MDC_OFFLINE = {"H":"/scratch/florent.robinet/BurstBenchmark/"
                                              ,"L":"/scratch/florent.robinet/BurstBenchmark/"
                                              ,"V":"/scratch/florent.robinet/BurstBenchmark/"}
                    )

    channels_dict = dict( DEFAULT_ONLINE = {"H":"" ,"L":"" ,"V":""}
    
                        , DEFAULT_OFFLINE = {"H":"" ,"L":"" ,"V":""}

                        , O4_ONLINE = {"H":"H1:GDS-CALIB_STRAIN_CLEAN"
                                      ,"L":"L1:GDS-CALIB_STRAIN_CLEAN"
                                      ,"V":"V1:Hrec_hoft_16384Hz_Gated"}

                        , O3MDC_ONLINE = {"H":"H1:GDS-CALIB_STRAIN_INJ1_O3Replay"
                                  ,"L":"L1:GDS-CALIB_STRAIN_INJ1_O3Replay"
                                  ,"V":"V1:Hrec_hoft_16384Hz_INJ1_O3Replay"}

                        , O3REPLAY_ONLINE = {"H":"H1:GDS-CALIB_STRAIN_O3Replay"
                                     ,"L":"L1:GDS-CALIB_STRAIN_O3Replay"
                                     ,"V":"V1:Hrec_hoft_16384Hz_INJ1_O3Replay"}

                        , BENCHMARK_MDC_OFFLINE = {"H":"H1:STRAIN_BURST_0"
                                                  ,"L":"L1:STRAIN_BURST_0"
                                                  ,"V":"V1:STRAIN_BURST_0"}
                                )
    
    state_vector_dict = dict( 
                          DEFAULT_ONLINE = {}

                        , DEFAULT_OFFLINE = {}

                        , O4_ONLINE = {"H":"H1:GDS-CALIB_STATE_VECTOR"
                                      ,"L":"L1:GDS-CALIB_STATE_VECTOR"
                                      ,"V":"V1:DQ_ANALYSIS_STATE_VECTOR"}
                        , O3MDC_ONLINE = {}
                        , O3REPLAY_ONLINE = {}
                        )

    active_flags_dict = dict( 
                          DEFAULT_ONLINE = {"H":"" ,"L":"" ,"V":""}

                        , DEFAULT_OFFLINE = {"H": "H1:DMT-ANALYSIS_READY:1"
                                            ,"L": "L1:DMT-ANALYSIS_READY:1"
                                            ,"V": "V1:ITF_SCIENCE:1"}

                        , O4_ONLINE = {"H": "H1:DMT-ANALYSIS_READY:1"
                                        ,"L": "L1:DMT-ANALYSIS_READY:1"
                                        ,"V": "V1:ITF_SCIENCE:1"}

                        , O3MDC_ONLINE = {"H": "H1:DMT-ANALYSIS_READY:1"
                                        ,"L": "L1:DMT-ANALYSIS_READY:1"
                                        ,"V": "V1:ITF_SCIENCE:1"}

                        , O3REPLAY_ONLINE = {"H": "H1:DMT-ANALYSIS_READY:1"
                                            ,"L": "L1:DMT-ANALYSIS_READY:1"
                                            ,"V": "V1:ITF_SCIENCE:1"}

                        , BENCHMARK_MDC_OFFLINE = {"H":"" ,"L":"" ,"V":""}
                                    )

                                

    if kwargs['search_mode'] == "BENCHMARK_MDC_OFFLINE":
        segment_list_default = "/home/vasileios.skliris/public_html/benchmark_segments.txt"
    else:
        segment_list_default = []

    if len(kwargs['detectors'])==3:
        skymap_config = dict(alpha = 2.2, beta = 1.0, sigma = 161.1 ,nside =64)
    elif len(kwargs['detectors'])==2:
        skymap_config = dict(alpha = 1.3, beta = 1.0, sigma = 35.9 ,nside =64)

    # # # # # # # CONFIG FILE # # # # # # #

    configuration = dict(
        

    # Mode that creates some presets if used
    search_mode  = kwargs["search_mode"],
    # # # Directory organisation 
    
    output_directory = kwargs["output_directory"],
    trigger_directory = kwargs["trigger_directory"],
    masterDirectory = kwargs["masterDirectory"],
    bufferDirectory = kwargs["bufferDirectory"],
    falseAlarmRates = kwargs["falseAlarmRates"],
    efficiencies = kwargs["efficiencies"],
    log = "log",
    # Logging level
    log_level = "INFO",
        
    # User name
    
    user_name = os.environ['HOME'].split("/")[-1],


    accounting_group_user = os.environ['HOME'].split("/")[-1],
    # Acounting Group
    accounting_group = "ligo.dev.o4.burst.allsky.mlyonline",

    # Search directory path
    path = os.path.abspath(kwargs['path']),
        
    # # # Generator Parameters    
    
    # Num samples per second to feed into models.
    fs=1024,           
        
    # Time duration of segments to analyse (analysis window).
    duration=1,         
        
    # Number of segments to analyse, keep to 1 for real-time analysis.
    size=1,
    
    # Detectors used for this search
    detectors = kwargs['detectors'],
    
    # Frame file prefixes to use
    frames_directory= frames_dict[kwargs['search_mode']],
    
    # The channels to use, INJ is the MDC and NOISE is O3-replay
    channels = channels_dict[kwargs['search_mode']],

    # The state vector used, used only in online searces
    state_vectors = state_vector_dict[kwargs['search_mode']],

    # The active flag used
    active_flags= active_flags_dict[kwargs['search_mode']],

    segment_list = segment_list_default,

    max_continuous_segment = 10000,
    # # # Fetching data related parameters 

    # The number of search script running. (The number of scripts must be more than 
    # the time it takes to process one instance (second) of data.)
    parallel_scripts=4,

    # Number of seconds to wait between each read attempt.
    wait=0.5,
        
    # Number of seconds required for PSD calculation.
    required_buffer=16,  
        
    # Number of seconds before current GPS time to begin search.
    start_lag=92,
        
    # If search time is left behind by that amount of seconds it skips ahead.
    gps_reset_time=32,
        
    # FAR file that is used to calculate FAR. It is used only at
    # the beggining of the search, until there is enough background estimation.
    
    farfile= ("/home/vasileios.skliris/mly-hermes/outputs/FARfile" if ('ONLINE' in kwargs['search_mode']) else kwargs["falseAlarmRates"]+"/FARfile" ),
    
    # # # Models 
        
    # Model1 path, conincidence model.
    model1_path=("/home/mly/models/model1_32V_No5.h5"), 
        
    # Model2 path, coherence model.
    model2_path=("/home/mly/models/model2_32V_No6.h5"),  

    # # # Skymap parameters
    
    # Skymap parameters
    skymap = skymap_config,
        
    # # # Efficiency calculation parametersp

    # Testing configuration
    eff_config = dict( injectionDirectoryPath="/home/mly/injections/" # Injection directories
                      ,injectionsWithHRSS = ['SGE70Q3'
                                            ,'SGE153Q8d9'
                                            ,'SGL153Q8d9'
                                            ,'WNB250'] if "OFFLINE" not in kwargs['search_mode'] else []
                      ,injectionsWithSNR = ['cbc_20_20'
                                             ,'wnb_03_train_pod'
                                             ,'cusp_00'] if "OFFLINE" not in kwargs['search_mode'] else []
                      ,injectionHRSS = "1e-22,1e-21,1e-22"
                      ,injectionSNR =  "0,50,5"
                      ,testSize = "100"
                      ,howOften = 3600),

    far_config = dict( 
                    # Enabling or not the calculation of FAR.
                    far_enabled = not kwargs['nofar']
                    # Number of timelags used.
                    ,batch_size=1024
                    # It is used only at the beggining of a search until a FAR is estimated.
                    ,threshold = 2.3148e-05 if "OFFLINE" in kwargs['search_mode'] else 1/3600
                    # The score from which we keep FAR statistics.
                    ,restriction = 0.0001
                    # The maximum lag to be used. Determines the DataSet size of the saved instances.
                    ,max_lag = 3600
                    # The amount of lags to use for each file. Depends on the capasity to produce them.
                    ,lags = 1024
                    # Then number of condor jobs per file to be used for timelags. With a file of
                    # size 3600 (max_lag) this will create jobs that will produce files
                    # of ~3600*lags/batches. Change with caution.
                    ,batches = 64
                    # GPU devices to use. If not defined it will just load the local visible ones.
                    ,visible_gpu_devices = "local"
                    # The GPU number(s)) (run nvidia-smi to see the numbers), to use for the tests.
                    ,selectedGPUs = [0]
                    # Maximum amount of dags to run at the same time in condor for the generation of data.
                    ,parallelGenerations = 3
                    ),
                         
    # # # MISC
    
    # The group size of the files that used to save the outputs    
    maxDataFrameSize = 3600,
    
    # The destination of the triggers, None means testing mode with FAKEIDs.
    trigger_destination = "playground" if kwargs['search_mode'] in ["O3MDC_ONLINE"] else None,
    # For the online search only
    trigger_group = "Burst" if 'ONLINE' in kwargs['search_mode'] else None,
    # Random number generator seed
    seed = 921101

    

    )


    os.system("mkdir " + configuration['path'] +"/"+ configuration['output_directory'])
    os.system("mkdir " + configuration['path'] +"/"+ configuration['trigger_directory'])
    os.system("mkdir " + configuration['path'] +"/"+ configuration['log'])

    os.system("mkdir " + configuration['path'] +"/"+ configuration['masterDirectory'])
    os.system("mkdir " + configuration['path'] +"/"+ configuration['masterDirectory'] +"/temp")
    for det in configuration['detectors']:
        os.system("mkdir " + configuration['path'] +"/"+ configuration['masterDirectory'] +"/"+ det)
        os.system("mkdir " + configuration['path'] +"/"+ configuration['masterDirectory'] +"/temp/"+ det)

    os.system("mkdir " + configuration['path'] +"/"+ configuration['bufferDirectory'])
    os.system("mkdir " + configuration['path'] +"/"+ configuration['bufferDirectory'] +"/temp")

    os.system("mkdir " + configuration['path'] +"/"+ configuration['efficiencies'])
    os.system("mkdir " + configuration['path'] +"/"+ configuration['efficiencies'] +"/history")

    if not kwargs['nofar']:

        os.system("mkdir " + configuration['path'] +"/"+ configuration['falseAlarmRates'])
        os.system("mkdir " + configuration['path'] +"/"+ configuration['falseAlarmRates']+"/hourly")
        os.system("mkdir " + configuration['path'] +"/"+ configuration['falseAlarmRates']+"/FARfile")
        os.system("mkdir " + configuration['path'] +"/"+ configuration['falseAlarmRates']+"/condor")
        os.system("mkdir " + configuration['path'] +"/"+ configuration['falseAlarmRates']+"/temp")
        
    # Saving config file to json format
    with open(configuration['path'] +"/"+"config.json", "w") as config_json:
        json.dump(configuration, config_json,indent=4)
        config_json.close()

    
    createRunAllScript(**configuration)

def _start_search():

    #Construct argument parser:
    parser = argparse.ArgumentParser()

    parser.add_argument('-p', '--search_path', type=str
                        , default = "./"
                        , help = "The path to the search to start"
                        , required = False)

    # Pass arguments:
    args = parser.parse_args()
    

    search_path = getattr(args, "search_path")      

    if os.path.isfile(search_path+"runall.sh"):

        os.system(f"cd {search_path} ; nohup bash runall.sh &> runall.out & echo $!>>.jobids.txt")
        print("Search has started!")

    else:

        raise FileNotFoundError(f"The path provided {search_path} does not have a runall.sh file to start the search.")

      
def _end_search():

    #Construct argument parser:
    parser = argparse.ArgumentParser()

    parser.add_argument('-p', '--search_path', type=str
                        , default = "./"
                        , help = "The path to the search to start"
                        , required = False)

    # Pass arguments:
    args = parser.parse_args()
    

    search_path = getattr(args, "search_path")      
    if os.path.isfile(search_path+".jobids.txt"):

        with open(f"{search_path}/.jobids.txt", "r") as idfile:
            idfile_ = idfile.read()
  
        # replacing end of line('/n') with ' ' and
        # splidfile.read()
    
        # replacing end of line('/n') with ' ' and
        # splitting the text it further when '.' is seen.
        job_id_list = idfile_.replace('\n', ' ').split(" ")
    
        for id_ in job_id_list:
            try:
                os.system(f"kill {str(id_)} >/dev/null 2>&1")#,stderr=subprocess.STDOUT)
            except Exception as e:
                pass

        try:
            os.system(f"singularity instance stop --all >/dev/null 2>&1")
        except Exception as e:
            pass

        os.system(f">{search_path}/.jobids.txt")

        print("Search stopped")
    else:

        raise FileNotFoundError(f"The path provided {search_path} does not have a .jobids.txt file to start the search.")


def _status_page():

    from .manager import manage_status_page
    manage_status_page()

def _search_status():

    #Construct argument parser:
    parser = argparse.ArgumentParser()

    parser.add_argument('-n', '--n', type=int
                        , default = "3"
                        , help = "The number of lines to show from the file."
                        , required = False)
    
    parser.add_argument('-p', '--search_path', type=str
                        , default = "./"
                        , help = "The path to the search to start"
                        , required = False)

    # Pass arguments:
    args = parser.parse_args()
    

    search_path = getattr(args, "search_path") 
    n = getattr(args, "n") 
    with open('monitor.log', 'r') as file:
        lines = file.readlines()
    for i, line in enumerate(lines[-n:], start=1):
        if i == n:
            print("\033[1m" + line.strip() + "\033[0m")  # Use ANSI escape codes for bold text
        else:
            print(line.strip())
    


def get_file_counts_after_createFileSystem():

    with open('config.json') as json_file:
        config = json.load(json_file)
    
    master_directory = config['masterDirectory']
    detectors = config['detectors']
    # Counting files after the createFileSystem
    file_counts = {}
    for detector in detectors:
        path = os.path.join(master_directory, detector)
        file_counts[detector] = len([f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))])

    common_files = set.intersection(*(set(os.listdir(os.path.join(master_directory, detector))) for detector in detectors))

    if len(set(file_counts.values())) == 1:
        print("All detectors have the same number of files. Successful jobs: ")
    else:
        # Deleting files that are not in all detectors.
        for detector in detectors:
            path = os.path.join(master_directory, detector)
            non_common_files = set(os.listdir(path)) - common_files
            for file_name in non_common_files:
                #print(os.path.join(path, file_name))
                os.remove(os.path.join(path, file_name))
        
        print("Some files not present in all directories, were removed. Successful jobs: ")

    final_file_counts = min([file_counts[detector] for detector in detectors])
    print(final_file_counts)


if __name__ == "__main__":
    
     main()

